# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from functools import partial
from typing import Callable, NamedTuple, TypeVar, Union

from jax import lax
from jax import numpy as jnp
from jax import random, tree_util
from jax.experimental import io_callback
from jax.scipy.special import expit

from .lax import cond, fori_loop, while_loop
from .tree_math import random_like

_DEBUG_FLAG = False

_DEBUG_TREE_END_IDXS = []
_DEBUG_SUBTREE_END_IDXS = []
_DEBUG_STORE = []


def _DEBUG_ADD_QP(qp):
    """Stores **all** results of leapfrog integration"""
    global _DEBUG_STORE
    _DEBUG_STORE.append(qp)


def _DEBUG_FINISH_TREE(dummy_arg):
    """Signal the position of a finished tree in `_DEBUG_STORE`"""
    global _DEBUG_TREE_END_IDXS
    _DEBUG_TREE_END_IDXS.append(len(_DEBUG_STORE))


def _DEBUG_FINISH_SUBTREE(dummy_arg):
    """Signal the position of a finished sub-tree in `_DEBUG_STORE`"""
    global _DEBUG_SUBTREE_END_IDXS
    _DEBUG_SUBTREE_END_IDXS.append(len(_DEBUG_STORE))


def select(pred, on_true, on_false):
    return tree_util.tree_map(partial(lax.select, pred), on_true, on_false)


### COMMON FUNCTIONALITY
Q = TypeVar("Q")


class QP(NamedTuple):
    """Object holding a pair of position and momentum.

    Attributes
    ----------
    position : Q
        Position.
    momentum : Q
        Momentum.
    """
    position: Q
    momentum: Q


def flip_momentum(qp: QP) -> QP:
    return QP(position=qp.position, momentum=-qp.momentum)


def sample_momentum_from_diagonal(*, key, mass_matrix_sqrt):
    """
    Draw a momentum sample from the kinetic energy of the hamiltonian.

    Parameters
    ----------
    key: ndarray
        PRNGKey used as the random key.
    mass_matrix_sqrt: ndarray
        The left square-root mass matrix (i.e. square-root of the inverse
        diagonal covariance) to use for sampling. Diagonal matrix represented
        as (possibly pytree of) ndarray vector containing the entries of the
        diagonal.
    """
    normal = random_like(key=key, primals=mass_matrix_sqrt, rng=random.normal)
    return tree_util.tree_map(jnp.multiply, mass_matrix_sqrt, normal)


# TODO: how to randomize step size (neal sect. 3.2)
# @partial(jit, static_argnames=('potential_energy_gradient',))
def leapfrog_step(
    potential_energy_gradient,
    kinetic_energy_gradient,
    step_size,
    inverse_mass_matrix,
    qp: QP,
):
    """
    Perform one iteration of the leapfrog integrator forwards in time.

    Parameters
    ----------
    potential_energy_gradient: Callable[[ndarray], float]
        Potential energy gradient part of the hamiltonian (V). Depends on
        position only.
    qp: QP
        Point in position and momentum space from which to start integration.
    step_size: float
        Step length (usually called epsilon) of the leapfrog integrator.
    """
    position = qp.position
    momentum = qp.momentum

    momentum_halfstep = (
        momentum - (step_size / 2.) * potential_energy_gradient(position)
    )

    position_fullstep = position + step_size * kinetic_energy_gradient(
        inverse_mass_matrix, momentum_halfstep
    )

    momentum_fullstep = (
        momentum_halfstep -
        (step_size / 2.) * potential_energy_gradient(position_fullstep)
    )

    qp_fullstep = QP(position=position_fullstep, momentum=momentum_fullstep)

    global _DEBUG_FLAG
    if _DEBUG_FLAG:
        # append result to global list variable
        io_callback(_DEBUG_ADD_QP, None, qp_fullstep)

    return qp_fullstep


### SIMPLE HMC
class AcceptedAndRejected(NamedTuple):
    accepted_qp: QP
    rejected_qp: QP
    accepted: Union[jnp.ndarray, bool]
    diverging: Union[jnp.ndarray, bool]


# @partial(jit, static_argnames=('potential_energy', 'potential_energy_gradient'))
def generate_hmc_acc_rej(
    *, key, initial_qp, potential_energy, kinetic_energy, inverse_mass_matrix,
    stepper, num_steps, step_size, max_energy_difference
) -> AcceptedAndRejected:
    """
    Generate a sample given the initial position.

    Parameters
    ----------
    key: ndarray
        a PRNGKey used as the random key
    position: ndarray
        The the starting position of this step of the markov chain.
    potential_energy: Callable[[ndarray], float]
        The potential energy, which is the distribution to be sampled from.
    mass_matrix: ndarray
        The mass matrix used in the kinetic energy
    num_steps: int
        The number of steps the leapfrog integrator should perform.
    step_size: float
        The step size (usually epsilon) for the leapfrog integrator.
    """
    loop_body = partial(stepper, step_size, inverse_mass_matrix)
    new_qp = fori_loop(
        lower=0,
        upper=num_steps,
        body_fun=lambda _, args: loop_body(args),
        init_val=initial_qp
    )
    # this flipping is needed to make the proposal distribution symmetric
    # doesn't have any effect on acceptance though because kinetic energy depends on momentum^2
    # might have an effect with other kinetic energies though
    proposed_qp = flip_momentum(new_qp)

    total_energy = partial(
        total_energy_of_qp,
        potential_energy=potential_energy,
        kinetic_energy_w_inv_mass=partial(kinetic_energy, inverse_mass_matrix)
    )
    energy_diff = total_energy(initial_qp) - total_energy(proposed_qp)
    energy_diff = jnp.where(jnp.isnan(energy_diff), jnp.inf, energy_diff)
    transition_probability = jnp.minimum(1., jnp.exp(energy_diff))

    accept = random.bernoulli(key, transition_probability)
    accepted_qp, rejected_qp = select(
        accept,
        (proposed_qp, initial_qp),
        (initial_qp, proposed_qp),
    )
    diverging = jnp.abs(energy_diff) > max_energy_difference
    return AcceptedAndRejected(
        accepted_qp, rejected_qp, accepted=accept, diverging=diverging
    )


### NUTS
class Tree(NamedTuple):
    """Object carrying tree metadata.

    Attributes
    ----------
    left, right : QP
        Respective endpoints of the trees path.
    logweight: Union[jnp.ndarray, float]
        Sum over all -H(q, p) in the tree's path.
    proposal_candidate: QP
        Sample from the trees path, distributed as exp(-H(q, p)).
    turning: Union[jnp.ndarray, bool]
        Indicator for either the left or right endpoint are a uturn or any
        subtree is a uturn.
    diverging: Union[jnp.ndarray, bool]
        Indicator for a large increase in energy in the next larger tree.
    depth: Union[jnp.ndarray, int]
        Levels of the tree.
    cumulative_acceptance: Union[jnp.ndarray, float]
        Sum of all acceptance probabilities relative to some initial energy
        value. This value is distinct from `logweight` as its absolute value is
        only well defined for the very final tree of NUTS.
    """
    left: QP
    right: QP
    logweight: Union[jnp.ndarray, float]
    proposal_candidate: QP
    turning: Union[jnp.ndarray, bool]
    diverging: Union[jnp.ndarray, bool]
    depth: Union[jnp.ndarray, int]
    cumulative_acceptance: Union[jnp.ndarray, float]


def total_energy_of_qp(qp, potential_energy, kinetic_energy_w_inv_mass):
    return potential_energy(qp.position
                           ) + kinetic_energy_w_inv_mass(qp.momentum)


def generate_nuts_tree(
    initial_qp,
    key,
    step_size,
    max_tree_depth,
    stepper: Callable[[Union[jnp.ndarray, float], Q, QP], QP],
    potential_energy,
    kinetic_energy: Callable[[Q, Q], float],
    inverse_mass_matrix: Q,
    bias_transition: bool = True,
    max_energy_difference: Union[jnp.ndarray, float] = jnp.inf
) -> Tree:
    """Generate a sample given the initial position.

    This call implements a No-U-Turn-Sampler.

    Parameters
    ----------
    initial_qp: QP
        Starting pair of (position, momentum). **NOTE**, the momentum must be
        resampled from conditional distribution **BEFORE** passing it into this
        function!
    key: ndarray
        PRNGKey used as the random key.
    step_size: float
        Step size (usually called epsilon) for the leapfrog integrator.
    max_tree_depth: int
        The maximum depth of the trajectory tree before the expansion is
        terminated. At the maximum iteration depth, the current value is
        returned even if the U-turn condition is not met. The maximum number of
        points (/integration steps) per trajectory is :math:`N =
        2^{\\mathrm{max\\_tree\\_depth}}`. This function requires memory linear
        in max_tree_depth, i.e. logarithmic in trajectory length. It is used to
        statically allocate memory in advance.
    stepper: Callable[[float, Q, QP], QP]
        The function that performs (Leapfrog) steps. Takes as arguments (in order)
        1) step size (containing the direction): float ,
        2) inverse mass matrix: Q ,
        3) starting point: QP .
    potential_energy: Callable[[Q], float]
        The potential energy, of the distribution to be sampled from. Takes
        only the position part (QP.position) as argument.
    kinetic_energy: Callable[[Q, Q], float], optional
        Mapping of the momentum to its corresponding kinetic energy. As
        argument the function takes the inverse mass matrix and the momentum.

    Returns
    -------
    current_tree: Tree
        The final tree, carrying a sample from the target distribution.

    See Also
    --------
    No-U-Turn Sampler original paper (2011): https://arxiv.org/abs/1111.4246
    NumPyro Iterative NUTS paper: https://arxiv.org/abs/1912.11554
    Combination of samples from two trees, Sampling from trajectories according to target distribution in this paper's Appendix: https://arxiv.org/abs/1701.02434
    """
    # initialize depth 0 tree, containing 2**0 = 1 points
    initial_neg_energy = -total_energy_of_qp(
        initial_qp, potential_energy,
        partial(kinetic_energy, inverse_mass_matrix)
    )
    current_tree = Tree(
        left=initial_qp,
        right=initial_qp,
        logweight=initial_neg_energy,
        proposal_candidate=initial_qp,
        turning=False,
        diverging=False,
        depth=0,
        cumulative_acceptance=jnp.zeros_like(initial_neg_energy)
    )

    def _cont_cond(loop_state):
        _, current_tree, stop = loop_state
        return (~stop) & (current_tree.depth <= max_tree_depth)

    def cond_tree_doubling(loop_state):
        key, current_tree, _ = loop_state
        key, key_dir, key_subtree, key_merge = random.split(key, 4)

        go_right = random.bernoulli(key_dir, 0.5)

        # build tree adjacent to current_tree
        new_subtree = iterative_build_tree(
            key_subtree,
            current_tree,
            step_size,
            go_right,
            stepper,
            potential_energy,
            kinetic_energy,
            inverse_mass_matrix,
            max_tree_depth,
            initial_neg_energy=initial_neg_energy,
            max_energy_difference=max_energy_difference
        )
        # Mark current tree as diverging if it diverges in the next step
        current_tree = current_tree._replace(diverging=new_subtree.diverging)

        # combine current_tree and new_subtree into a tree which is one layer deeper only if new_subtree has no turning subtrees (including itself)
        current_tree = cond(
            # If new tree is turning or diverging, do not merge
            pred=new_subtree.turning | new_subtree.diverging,
            true_fun=lambda old_and_new: old_and_new[0],
            false_fun=lambda old_and_new: merge_trees(
                key_merge,
                old_and_new[0],
                old_and_new[1],
                go_right,
                bias_transition=bias_transition
            ),
            operand=(current_tree, new_subtree),
        )
        # stop if new subtree was turning -> we sample from the old one and don't expand further
        # stop if new total tree is turning -> we sample from the combined trajectory and don't expand further
        stop = new_subtree.turning | current_tree.turning
        stop |= new_subtree.diverging
        return (key, current_tree, stop)

    loop_state = (key, current_tree, False)
    _, current_tree, _ = while_loop(_cont_cond, cond_tree_doubling, loop_state)

    global _DEBUG_FLAG
    if _DEBUG_FLAG:
        io_callback(_DEBUG_FINISH_TREE, None, None)

    return current_tree


def tree_index_get(ptree, idx):
    return tree_util.tree_map(lambda arr: arr[idx], ptree)


def tree_index_update(x, idx, y):
    from jax.tree_util import tree_map

    return tree_map(lambda x_el, y_el: x_el.at[idx].set(y_el), x, y)


def count_trailing_ones(n):
    """Count the number of trailing, consecutive ones in the binary
    representation of `n`.

    Warning
    -------
    `n` must be positive and strictly smaller than 2**64

    Examples
    --------
    >>> print(bin(23), count_trailing_one_bits(23))
    0b10111 3
    """
    # taken from http://num.pyro.ai/en/stable/_modules/numpyro/infer/hmc_util.html
    _, trailing_ones_count = while_loop(
        lambda nc: (nc[0] & 1) != 0, lambda nc: (nc[0] >> 1, nc[1] + 1), (n, 0)
    )
    return trailing_ones_count


def is_euclidean_uturn(qp_left, qp_right):
    """
    See Also
    --------
    Betancourt - A conceptual introduction to Hamiltonian Monte Carlo
    """
    return (
        (qp_right.momentum.dot(qp_right.position - qp_left.position) < 0.) &
        (qp_left.momentum.dot(qp_left.position - qp_right.position) < 0.)
    )


# Essentially algorithm 2 from https://arxiv.org/pdf/1912.11554.pdf
def iterative_build_tree(
    key, initial_tree, step_size, go_right, stepper, potential_energy,
    kinetic_energy, inverse_mass_matrix, max_tree_depth, initial_neg_energy,
    max_energy_difference
):
    """
    Starting from either the left or right endpoint of a given tree, builds a
    new adjacent tree of the same size.

    Parameters
    ----------
    key: ndarray
        PRNGKey to choose a sample when adding QPs to the tree.
    initial_tree: Tree
        Tree to be extended (doubled) on the left or right.
    step_size: float
        The step size (usually called epsilon) for the leapfrog integrator.
    go_right: bool
        If `go_right` start at the right end, going right else start at the
        left end, going left.
    stepper: Callable[[float, Q, QP], QP]
        The function that performs (Leapfrog) steps. Takes as arguments (in order)
        1) step size (containing the direction): float ,
        2) inverse mass matrix: Q ,
        3) starting point: QP .
    potential_energy: Callable[[Q], float]
        Potential energy, of the distribution to be sampled from. Takes
        only the position part (QP.position) as argument.
    kinetic_energy: Callable[[Q, Q], float], optional
        Mapping of the momentum to its corresponding kinetic energy. As
        argument the function takes the inverse mass matrix and the momentum.
    max_tree_depth: int
        An upper bound on the 'depth' argument, but has no effect on the
        functions behaviour. It's only required to statically set the size of
        the `S` array (Q).
    """
    # 1. choose start point of integration
    z = select(go_right, initial_tree.right, initial_tree.left)
    depth = initial_tree.depth
    max_num_proposals = 2**depth
    # 2. build / collect new states
    # Create a storage for left endpoints of subtrees. Size is determined
    # statically by the `max_tree_depth` parameter.
    # NOTE, let's hope this does not break anything but in principle we only
    # need `max_tree_depth` element even though the tree can be of length `max_tree_depth +
    # 1`. This is because we will never access the last element.
    S = tree_util.tree_map(
        lambda proto: jnp.
        empty_like(proto, shape=(max_tree_depth, ) + jnp.shape(proto)), z
    )

    z = stepper(
        jnp.where(go_right, 1., -1.) * step_size, inverse_mass_matrix, z
    )
    neg_energy = -total_energy_of_qp(
        z, potential_energy, partial(kinetic_energy, inverse_mass_matrix)
    )
    diverging = jnp.abs(neg_energy - initial_neg_energy) > max_energy_difference
    cum_acceptance = jnp.minimum(1., jnp.exp(initial_neg_energy - neg_energy))
    incomplete_tree = Tree(
        left=z,
        right=z,
        logweight=neg_energy,
        proposal_candidate=z,
        turning=False,
        diverging=diverging,
        depth=-1,
        cumulative_acceptance=cum_acceptance
    )
    S = tree_index_update(S, 0, z)

    def amend_incomplete_tree(state):
        n, incomplete_tree, z, S, key = state

        key, key_choose_candidate = random.split(key)
        z = stepper(
            jnp.where(go_right, 1., -1.) * step_size, inverse_mass_matrix, z
        )
        incomplete_tree = add_single_qp_to_tree(
            key_choose_candidate,
            incomplete_tree,
            z,
            go_right,
            potential_energy,
            kinetic_energy,
            inverse_mass_matrix,
            initial_neg_energy=initial_neg_energy,
            max_energy_difference=max_energy_difference
        )

        def _even_fun(S):
            # n is even, the current z is w.l.o.g. a left endpoint of some
            # subtrees. Register the current z to be used in turning condition
            # checks later, when the right endpoints of it's subtrees are
            # generated.
            S = tree_index_update(S, lax.population_count(n), z)
            return S, False

        def _odd_fun(S):
            # n is odd, the current z is w.l.o.g a right endpoint of some
            # subtrees. Check turning condition against all left endpoints of
            # subtrees that have the current z (/n) as their right endpoint.

            # l = nubmer of subtrees that have current z as their right endpoint.
            l = count_trailing_ones(n)
            # inclusive indices into S referring to the left endpoints of the l subtrees.
            i_max_incl = lax.population_count(n - 1)
            i_min_incl = i_max_incl - l + 1
            # TODO: this should traverse the range in reverse
            turning = fori_loop(
                lower=i_min_incl,
                upper=i_max_incl + 1,
                # TODO: conditional for early termination
                body_fun=lambda k, turning: turning |
                is_euclidean_uturn(tree_index_get(S, k), z),
                init_val=False
            )
            return S, turning

        S, turning = cond(
            pred=n % 2 == 0, true_fun=_even_fun, false_fun=_odd_fun, operand=S
        )
        incomplete_tree = incomplete_tree._replace(turning=turning)
        return (n + 1, incomplete_tree, z, S, key)

    def _cont_cond(state):
        n, incomplete_tree, *_ = state
        return (n < max_num_proposals) & (~incomplete_tree.turning
                                         ) & (~incomplete_tree.diverging)

    n, incomplete_tree, *_ = while_loop(
        # while n < 2**depth and not stop
        cond_fun=_cont_cond,
        body_fun=amend_incomplete_tree,
        init_val=(1, incomplete_tree, z, S, key)
    )

    global _DEBUG_FLAG
    if _DEBUG_FLAG:
        io_callback(_DEBUG_FINISH_SUBTREE, None, None)

    # The depth of a tree which was aborted early is possibly ill defined
    depth = jnp.where(n == max_num_proposals, depth, -1)
    return incomplete_tree._replace(depth=depth)


def add_single_qp_to_tree(
    key, tree, qp, go_right, potential_energy, kinetic_energy,
    inverse_mass_matrix, initial_neg_energy, max_energy_difference
):
    """Helper function for progressive sampling. Takes a tree with a sample, and
    a new endpoint, propagates sample.
    """
    # This is technically just a special case of merge_trees with one of the
    # trees being a singleton, depth 0 tree. However, no turning check is
    # required and it is not possible to bias the transition.
    left, right = select(go_right, (tree.left, qp), (qp, tree.right))

    neg_energy = -total_energy_of_qp(
        qp, potential_energy, partial(kinetic_energy, inverse_mass_matrix)
    )
    diverging = jnp.abs(neg_energy - initial_neg_energy) > max_energy_difference
    # ln(e^-H_1 + e^-H_2)
    total_logweight = jnp.logaddexp(tree.logweight, neg_energy)
    # expit(x-y) := 1 / (1 + e^(-(x-y))) = 1 / (1 + e^(y-x)) = e^x / (e^y + e^x)
    prob_of_keeping_old = expit(tree.logweight - neg_energy)
    remain = random.bernoulli(key, prob_of_keeping_old)
    proposal_candidate = select(remain, tree.proposal_candidate, qp)
    # NOTE, set an invalid depth as to indicate that adding a single QP to a
    # perfect binary tree does not yield another perfect binary tree
    cum_acceptance = tree.cumulative_acceptance + jnp.minimum(
        1., jnp.exp(initial_neg_energy - neg_energy)
    )
    return Tree(
        left,
        right,
        total_logweight,
        proposal_candidate,
        turning=tree.turning,
        diverging=diverging,
        depth=-1,
        cumulative_acceptance=cum_acceptance
    )


def merge_trees(key, current_subtree, new_subtree, go_right, bias_transition):
    """Merges two trees, propagating the proposal_candidate"""
    # 5. decide which sample to take based on total weights (merge trees)
    if bias_transition:
        # Bias the transition towards the new subtree (see Betancourt
        # conceptual intro (and Numpyro))
        transition_probability = jnp.minimum(
            1., jnp.exp(new_subtree.logweight - current_subtree.logweight)
        )
    else:
        # expit(x-y) := 1 / (1 + e^(-(x-y))) = 1 / (1 + e^(y-x)) = e^x / (e^y + e^x)
        transition_probability = expit(
            new_subtree.logweight - current_subtree.logweight
        )
    # print(f"prob of choosing new sample: {transition_probability}")
    new_sample = select(
        random.bernoulli(key, transition_probability),
        new_subtree.proposal_candidate, current_subtree.proposal_candidate
    )
    # 6. define new tree
    left, right = select(
        go_right,
        (current_subtree.left, new_subtree.right),
        (new_subtree.left, current_subtree.right),
    )
    turning = is_euclidean_uturn(left, right)
    diverging = current_subtree.diverging | new_subtree.diverging
    neg_energy = jnp.logaddexp(new_subtree.logweight, current_subtree.logweight)
    cum_acceptance = current_subtree.cumulative_acceptance + new_subtree.cumulative_acceptance
    merged_tree = Tree(
        left=left,
        right=right,
        logweight=neg_energy,
        proposal_candidate=new_sample,
        turning=turning,
        diverging=diverging,
        depth=current_subtree.depth + 1,
        cumulative_acceptance=cum_acceptance
    )
    return merged_tree



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from functools import partial

import jax
import pytest
from jax import numpy as jnp
from jax import random
from jax.tree_util import tree_map
from numpy.testing import assert_allclose

import nifty8.re as jft
from nifty8.re.likelihood import partial_insert_and_remove as jpartial

pmp = pytest.mark.parametrize


def _identity(x):
    return x


def tree_assert_allclose(a, b, **kwargs):
    return tree_map(partial(assert_allclose, **kwargs), a, b)


def test_partial_insert_and_remove():
    # _identity takes exactly one argument, thus `insert_axes` and `flat_fill`
    # are length one tuples
    _id_part = jpartial(
        _identity,
        insert_axes=(jft.Vector({
            "a": (True, False),
            "b": False
        }), ),
        flat_fill=(("THIS IS input['a'][0]", ), )
    )
    out = _id_part(("THIS IS input['a'][1]", "THIS IS input['b']"))
    assert out == jft.Vector(
        {
            "a": ("THIS IS input['a'][0]", "THIS IS input['a'][1]"),
            "b": "THIS IS input['b']"
        }
    )


def _identity(x):
    return x


@pmp("seed", (33, 42, 43))
@pmp("forward", (_identity, jnp.exp))
def test_likelihood_partial(seed, forward):
    atol, rtol = 1e-14, 1e-14
    aallclose = partial(assert_allclose, rtol=rtol, atol=atol)

    key = random.PRNGKey(seed)
    forward = partial(tree_map, forward)

    domain = jft.Vector(
        {
            "a": jft.ShapeWithDtype(128),
            "b": jft.ShapeWithDtype(64)
        }
    )
    key, sk_d, sk_p = random.split(key, 3)
    primals = jft.random_like(sk_p, domain)
    data = forward(jft.random_like(sk_d, domain))

    gaussian = jft.Gaussian(data).amend(forward)

    gaussian_part, primals_liquid = gaussian.freeze(
        primals=primals, point_estimates=("b", )
    )
    assert primals_liquid.tree[0].shape == domain["a"].shape
    aallclose(gaussian_part(primals_liquid), gaussian(primals))
    tree_map(
        aallclose,
        gaussian_part.left_sqrt_metric(primals_liquid, data).tree[0],
        gaussian.left_sqrt_metric(primals, data).tree["a"],
    )
    rsm_orig = gaussian.right_sqrt_metric(primals, primals)
    rsm_orig.tree["b"] = jft.zeros_like(rsm_orig.tree["b"])
    tree_map(
        aallclose,
        gaussian_part.right_sqrt_metric(primals_liquid, primals_liquid),
        rsm_orig,
    )
    aallclose(
        gaussian_part.metric(primals_liquid, primals_liquid).tree[0],
        gaussian.metric(primals, primals).tree["a"],
    )


def test_likelihood_domain():
    lh = jft.Poissonian(jnp.zeros((3, 4), dtype=int))
    assert lh.domain is not None
    lh_wfwd = lh.amend(
        jft.Model(lambda x: x, domain=lh.domain, white_init=True)
    )
    assert lh_wfwd.domain is not None


def _Poissonian(x):
    return jft.Poissonian(tree_map(lambda y: jnp.exp(y).astype(int), x))


@pmp("seed", (33, 42, 43))
@pmp("likelihood", (jft.Gaussian, partial(jft.StudentT, dof=2.2), _Poissonian))
@pmp("forward_a,forward_b", ((_identity, jnp.exp), (jnp.exp, jnp.reciprocal)))
def test_nonvariable_likelihood_add(seed, likelihood, forward_a, forward_b):
    key = random.PRNGKey(seed)
    N_TRIES = 100
    shp_a, shp_b = (3, 5), (12, 5)
    swd_a = jft.Vector((jft.ShapeWithDtype(shp_a), ) * 2)
    swd_b = jft.Vector((jft.ShapeWithDtype(shp_b), ) * 3)
    key_a, key_b = "a", "b"

    def fwd_a(x):
        x = tree_map(forward_a, x[key_a])
        everything_pos = likelihood is _Poissonian
        return tree_map(jnp.abs, x) if everything_pos else x

    def fwd_b(x):
        x = tree_map(forward_b, x[key_b])
        everything_pos = likelihood is _Poissonian
        return tree_map(jnp.abs, x) if everything_pos else x

    def forward(x):
        return jft.Vector({key_a: fwd_a(x), key_b: fwd_b(x)})

    key, k_a, k_b = random.split(key, 3)
    data_a = jft.random_like(k_a, swd_a)
    data_b = jft.random_like(k_b, swd_b)
    data_ab = jft.Vector({key_a: data_a, key_b: data_b})

    swd = jft.Vector({key_a: swd_a, key_b: swd_b})
    lh_orig = likelihood(data_ab).amend(forward, domain=swd)
    lh_a = likelihood(data_a).amend(fwd_a, domain=jft.Vector({key_a: swd_a}))
    lh_b = likelihood(data_b).amend(fwd_b, domain=jft.Vector({key_b: swd_b}))
    lh_ab = lh_a + lh_b

    key, k_p, k_t, k_q = random.split(key, 4)
    rl = jax.vmap(jft.random_like, in_axes=(0, None))
    p, t, q = tuple(
        rl(random.split(k, N_TRIES), lh_orig.domain) for k in (k_p, k_t, k_q)
    )

    assert_allclose(jax.vmap(lh_orig)(p), jax.vmap(lh_ab)(p), equal_nan=False)
    rsm_orig = jax.vmap(lh_orig.right_sqrt_metric)(p, t)
    rsm_ab = jax.vmap(lh_ab.right_sqrt_metric)(p, t)
    tree_assert_allclose(
        rsm_orig.tree[key_a], rsm_ab["lh_left"], equal_nan=False
    )
    tree_assert_allclose(
        rsm_orig.tree[key_b], rsm_ab["lh_right"], equal_nan=False
    )
    tree_assert_allclose(
        jax.vmap(
            lambda p, t, q: lh_orig.
            left_sqrt_metric(p, lh_orig.right_sqrt_metric(t, q))
        )(p, t, q),
        jax.vmap(
            lambda p, t, q: lh_ab.
            left_sqrt_metric(p, lh_ab.right_sqrt_metric(t, q))
        )(p, t, q),
        equal_nan=False
    )
    tree_assert_allclose(
        jax.vmap(lh_orig.metric)(p, t),
        jax.vmap(lh_ab.metric)(p, t),
        equal_nan=False
    )
    nresi_orig = jax.vmap(lh_orig.normalized_residual)(p)
    nresi_ab = jax.vmap(lh_ab.normalized_residual)(p)
    tree_assert_allclose(
        nresi_orig.tree[key_a], nresi_ab["lh_left"], equal_nan=False
    )
    tree_assert_allclose(
        nresi_orig.tree[key_b], nresi_ab["lh_right"], equal_nan=False
    )
    trafo_orig = jax.vmap(lh_orig.transformation)(p)
    trafo_ab = jax.vmap(lh_ab.transformation)(p)
    tree_assert_allclose(
        trafo_orig.tree[key_a], trafo_ab["lh_left"], equal_nan=False
    )
    tree_assert_allclose(
        trafo_orig.tree[key_b], trafo_ab["lh_right"], equal_nan=False
    )


@pmp("seed", (33, 42, 43))
@pmp(
    "likelihood", (
        jft.VariableCovarianceGaussian,
        partial(jft.VariableCovarianceStudentT, dof=1.7),
    )
)
@pmp("forward_a,forward_b", ((_identity, jnp.exp), (jnp.exp, jnp.reciprocal)))
def test_variable_likelihood_add(seed, likelihood, forward_a, forward_b):
    key = random.PRNGKey(seed)
    N_TRIES = 100
    # Make input always a tuple for Variable* likelihoods
    data_shp_a = (3, 5)
    data_shp_b = (12, 5)
    primals_swd_a = (jft.ShapeWithDtype(data_shp_a), ) * 2
    data_swd_a = jft.ShapeWithDtype(data_shp_a)
    primals_swd_b = (jft.ShapeWithDtype(data_shp_b), ) * 2
    data_swd_b = jft.ShapeWithDtype(data_shp_b)
    key_a, key_b = "a", "b"

    def fwd_a(x):
        x1, x2 = tree_map(forward_a, x[key_a])
        return (x1, tree_map(jnp.abs, x2))

    def fwd_b(x):
        x1, x2 = tree_map(forward_b, x[key_b])
        return (x1, tree_map(jnp.abs, x2))

    def forward(x):
        a1, a2 = fwd_a(x)
        b1, b2 = fwd_b(x)
        x1 = jft.Vector({key_a: a1, key_b: b1})
        x2 = jft.Vector({key_a: a2, key_b: b2})
        return jft.Vector((x1, x2))

    key, k_a, k_b = random.split(key, 3)
    data_a = jft.random_like(k_a, data_swd_a)
    data_b = jft.random_like(k_b, data_swd_b)
    data_ab = jft.Vector({key_a: data_a, key_b: data_b})

    swd = jft.Vector({key_a: primals_swd_a, key_b: primals_swd_b})
    lh_orig = likelihood(data_ab).amend(forward, domain=swd)
    lh_a = likelihood(data_a).amend(
        fwd_a, domain=jft.Vector({key_a: primals_swd_a})
    )
    lh_b = likelihood(data_b).amend(
        fwd_b, domain=jft.Vector({key_b: primals_swd_b})
    )
    lh_ab = lh_a + lh_b

    key, k_p, k_t, k_q = random.split(key, 4)
    rl = jax.vmap(jft.random_like, in_axes=(0, None))
    p, t, q = tuple(
        rl(random.split(k, N_TRIES), lh_orig.domain) for k in (k_p, k_t, k_q)
    )

    assert_allclose(jax.vmap(lh_orig)(p), jax.vmap(lh_ab)(p), equal_nan=False)
    rsm_orig = jax.vmap(lh_orig.right_sqrt_metric)(p, t)
    rsm_ab = jax.vmap(lh_ab.right_sqrt_metric)(p, t)
    tree_assert_allclose(
        tuple(r.tree[key_a] for r in rsm_orig),
        rsm_ab["lh_left"],
        equal_nan=False
    )
    tree_assert_allclose(
        tuple(r.tree[key_b] for r in rsm_orig),
        rsm_ab["lh_right"],
        equal_nan=False
    )
    tree_assert_allclose(
        jax.vmap(
            lambda p, t, q: lh_orig.
            left_sqrt_metric(p, lh_orig.right_sqrt_metric(t, q))
        )(p, t, q),
        jax.vmap(
            lambda p, t, q: lh_ab.
            left_sqrt_metric(p, lh_ab.right_sqrt_metric(t, q))
        )(p, t, q),
        equal_nan=False
    )
    tree_assert_allclose(
        jax.vmap(lh_orig.metric)(p, t),
        jax.vmap(lh_ab.metric)(p, t),
        equal_nan=False
    )
    nresi_orig = lh_orig.normalized_residual(p)
    nresi_ab = lh_ab.normalized_residual(p)
    tree_assert_allclose(
        nresi_orig[key_a], nresi_ab["lh_left"], equal_nan=False
    )
    tree_assert_allclose(
        nresi_orig[key_b], nresi_ab["lh_right"], equal_nan=False
    )

    try:
        jax.vmap(lh_orig.transformation)(p)
    except NotImplementedError:
        return
    trafo_orig = jax.vmap(lh_orig.transformation)(p)
    trafo_ab = jax.vmap(lh_ab.transformation)(p)
    tree_assert_allclose(
        tuple(t[key_a] for t in trafo_orig),
        trafo_ab["lh_left"],
        equal_nan=False
    )
    tree_assert_allclose(
        tuple(t[key_b] for t in trafo_orig),
        trafo_ab["lh_right"],
        equal_nan=False
    )


if __name__ == "__main__":
    test_likelihood_partial(33, jnp.exp)
    test_nonvariable_likelihood_add(42, jft.Gaussian, jnp.exp, jnp.reciprocal)
    test_variable_likelihood_add(
        42, jft.VariableCovarianceGaussian, jnp.exp, jnp.reciprocal
    )
    test_nonvariable_likelihood_add(42, _Poissonian, jnp.exp, jnp.reciprocal)



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

_config = dict(
    hartley_convention="non_canonical_hartley",
)


def update(key, value, /):
    """Update the global configuration of NIFTy and NIFTy.re

    Parameters
    ----------
    key : str
        Identifier for the configuration option.
    value : Any
        Value for the configuration option.


    Currently, the following configuration options are available:

    - "hartley_convention": one of "non_canonical_hartley" or
      "canonical_hartley" for ducc's old non-canonical Hartley convention
      respectively ducc's new canononical Hartley convention
    """
    global _config
    if not isinstance(key, str):
        raise TypeError(f"key must be a string; got {key!r}")
    key = key.lower()
    if key == "hartley_convention":
        if not isinstance(value, str):
            raise TypeError(f"value to {key!r} must be a string; got {value!r}")
        if value in ("ducc_hartley", "non_canonical_hartley"):
            value = "non_canonical_hartley"
        elif value in ("ducc_fht", "canonical_hartley"):
            value = "canonical_hartley"
        else:
            raise ValueError(f"invalid value to {key!r}; got {value!r}")
    else:
        raise ValueError(f"invalid key; got {key!r}")
    _config[key] = value



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
# Authors: Gordian Edenhofer
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from nifty8.extra import check_linear_operator
from numpy.testing import assert_allclose, assert_array_equal

from ..common import list2fixture, setup_function, teardown_function

pmp = pytest.mark.parametrize

# The test cases do not work on a multi-dimensional RGSpace yet
spaces = (
    ift.UnstructuredDomain(4),
    ift.LMSpace(5),
    ift.GLSpace(4),
)
space1 = list2fixture(spaces)
space2 = list2fixture(spaces)
dtype = list2fixture([np.float64, np.complex128])


def test_split_operator_first_axes_without_intersections(
    space1, space2, n_splits=3
):
    rng = ift.random.current_rng()

    dom = ift.DomainTuple.make((space1, space2))
    orig_idx = np.arange(space1.shape[0])
    rng.shuffle(orig_idx)
    split_idx = np.array_split(orig_idx, n_splits)
    split = ift.SplitOperator(
        dom, {f"{i:06d}": (si, )
              for i, si in enumerate(split_idx)}
    )
    assert check_linear_operator(split) is None

    r = ift.from_random(dom, "normal")
    split_r = split(r)
    # This relies on the keys of the target domain either being in the order of
    # insertion or being alphabetically sorted
    for idx, v in zip(split_idx, split_r.val.values()):
        assert_array_equal(r.val[idx], v)
    # Here, the adjoint must be the inverse as the field is split fully among
    # the generated indices and without intersections.
    assert_array_equal(split.adjoint(split_r).val, r.val)


def test_split_operator_first_axes_with_intersections(
    space1, space2, n_splits=3
):
    rng = ift.random.current_rng()

    dom = ift.DomainTuple.make((space1, space2))
    orig_idx = np.arange(space1.shape[0])
    split_idx = [
        rng.choice(orig_idx, rng.integers(1, space1.shape[0]), replace=False)
        for _ in range(n_splits)
    ]
    split = ift.SplitOperator(
        dom, {f"{i:06d}": (si, )
              for i, si in enumerate(split_idx)}
    )
    print(split_idx)
    assert check_linear_operator(split) is None

    r = ift.from_random(dom, "normal")
    split_r = split(r)
    # This relies on the keys of the target domain either being in the order of
    # insertion or being alphabetically sorted
    for idx, v in zip(split_idx, split_r.val.values()):
        assert_array_equal(r.val[idx], v)

    r_diy = np.copy(r.val)
    unique_freq = np.unique(np.concatenate(split_idx), return_counts=True)
    # Null values that were not selected
    r_diy[list(set(unique_freq[0]) ^ set(range(space1.shape[0])))] = 0.
    for idx, freq in zip(*unique_freq):
        r_diy[idx] *= freq
    assert_allclose(split.adjoint(split_r).val, r_diy)



import numpy as np
import pytest
from numpy.testing import assert_allclose

from nifty8.ducc_dispatch import (
    _scipy_fftn, _scipy_hartley, _scipy_ifftn, _scipy_vdot, fftn, hartley,
    ifftn, vdot
)

pytest.importorskip("ducc0")

pmp = pytest.mark.parametrize


@pmp("seed", [42])
@pmp("shape", [(10), (32, 32), (16, 16, 16)])
@pmp("dtype", [np.float64, np.complex128])
def test_ffts(seed, shape, dtype):
    rng = np.random.default_rng(seed)
    array = rng.normal(0, 1, shape).astype(dtype, copy=False)
    assert_allclose(fftn(array), _scipy_fftn(array))


@pmp("seed", [42])
@pmp("shape", [(10), (32, 32), (16, 16, 16)])
@pmp("dtype", [np.float64, np.complex128])
def test_iffts(seed, shape, dtype):
    rng = np.random.default_rng(seed)
    array = rng.normal(0, 1, shape).astype(dtype, copy=False)
    assert_allclose(ifftn(array), _scipy_ifftn(array))


@pmp("seed", [42])
@pmp("shape", [(10), (32, 32), (16, 16, 16)])
@pmp("dtype", [np.float64])
def test_hartleys(seed, shape, dtype):
    rng = np.random.default_rng(seed)
    array = rng.normal(0, 1, shape).astype(dtype, copy=False)
    assert_allclose(hartley(array), _scipy_hartley(array))


@pmp("seed", [42])
@pmp("shape", [(10), (32, 32), (16, 16, 16)])
@pmp("dtype", [np.float64, np.complex128])
def test_vdots(seed, shape, dtype):
    rng = np.random.default_rng(seed)
    a = rng.normal(0, 1, shape).astype(dtype, copy=False)
    b = rng.normal(0, 1, shape).astype(dtype, copy=False)
    assert_allclose(vdot(a, b), _scipy_vdot(a, b))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_allclose

from ..common import list2fixture, setup_function, teardown_function


def _get_rtol(tp):
    if (tp == np.float64) or (tp == np.complex128):
        return 1e-10
    else:
        return 1e-5


tp = list2fixture([np.float64, np.float32, np.complex64, np.complex128])
lm = list2fixture([128, 256])
pmp = pytest.mark.parametrize


def test_dotsht(lm, tp):
    tol = 10*_get_rtol(tp)
    a = ift.LMSpace(lmax=lm)
    b = ift.GLSpace(nlat=lm + 1)
    fft = ift.HarmonicTransformOperator(domain=a, target=b)
    inp = ift.Field.from_random(domain=a, random_type='normal', dtype=tp, std=1, mean=0)
    out = fft.times(inp)
    v1 = np.sqrt(out.s_vdot(out))
    v2 = np.sqrt(inp.s_vdot(fft.adjoint_times(out)))
    assert_allclose(v1, v2, rtol=tol, atol=tol)


def test_dotsht2(lm, tp):
    tol = 10*_get_rtol(tp)
    a = ift.LMSpace(lmax=lm)
    b = ift.HPSpace(nside=lm//2)
    fft = ift.HarmonicTransformOperator(domain=a, target=b)
    inp = ift.Field.from_random(domain=a, random_type='normal', dtype=tp, std=1, mean=0)
    out = fft.times(inp)
    v1 = np.sqrt(out.s_vdot(out))
    v2 = np.sqrt(inp.s_vdot(fft.adjoint_times(out)))
    assert_allclose(v1, v2, rtol=tol, atol=tol)


@pmp('space', [ift.LMSpace(lmax=30, mmax=25)])
def test_normalisation(space, tp):
    tol = 10*_get_rtol(tp)
    cospace = space.get_default_codomain()
    fft = ift.HarmonicTransformOperator(space, cospace)
    inp = ift.Field.from_random(domain=space, random_type='normal', dtype=tp, std=1, mean=2)
    out = fft.times(inp)
    zero_idx = tuple([0]*len(space.shape))
    assert_allclose(
        inp.val[zero_idx], out.s_integrate(), rtol=tol, atol=tol)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_allclose

from ..common import list2fixture, setup_function, teardown_function
from .test_adjoint import _h_spaces, _p_spaces, _pow_spaces

pmp = pytest.mark.parametrize


@pmp("dom0", _h_spaces + _p_spaces + _pow_spaces)
@pmp("dom1", _h_spaces + _p_spaces + _pow_spaces)
def test_multifield2vector(dom0, dom1):
    fld = ift.MultiField.from_dict({'a': ift.from_random(dom0),
                                    'b': ift.from_random(dom1)})
    op = ift.Multifield2Vector(fld.domain)
    ift.extra.assert_allclose(fld, op.adjoint(op(fld)))
    assert op.target.size == fld.size
    
    arr = np.empty(fld.size)
    arr[:dom0.size] = fld['a'].val.flatten()
    arr[dom0.size:] = fld['b'].val.flatten()
    assert_allclose(arr, op(fld).val)



# Copyright(C) 2013-2021 Max-Planck-Society
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import operator
from functools import partial, reduce
from typing import Any, Callable, Tuple, Union, Dict
from warnings import warn

from . import re as jft
from .domain_tuple import DomainTuple
from .field import Field
from .multi_domain import MultiDomain
from .multi_field import MultiField
from .operators.operator import Operator
from .sugar import makeField


def spaces_to_axes(domain, spaces) -> Union[Tuple, int, None]:
    """Converts spaces in a domain to axes of the underlying NumPy array."""
    if spaces is None:
        return None

    domain = DomainTuple.make(domain)
    axes = tuple(domain.axes[sp_index] for sp_index in spaces)
    axes = reduce(operator.add, axes) if len(axes) > 0 else axes
    return axes


def shapewithdtype_from_domain(
    domain, dtype
) -> Union[jft.ShapeWithDtype, Dict[str, jft.ShapeWithDtype]]:
    if isinstance(dtype, dict):
        dtp_fallback = float  # Fallback to `float` for unspecified keys
        k2dtp = dtype
    else:
        dtp_fallback = dtype
        k2dtp = {}

    if isinstance(domain, MultiDomain):
        parameter_tree = {}
        for k, dom in domain.items():
            parameter_tree[k] = jft.ShapeWithDtype(
                dom.shape, k2dtp.get(k, dtp_fallback)
            )
    elif isinstance(domain, DomainTuple):
        parameter_tree = jft.ShapeWithDtype(domain.shape, dtype)
    else:
        raise TypeError(f"incompatible domain {domain!r}")
    return parameter_tree


def wrap_nifty_call(op, target_dtype=float) -> Callable[[Any], jft.Vector]:
    from jax import pure_callback

    if callable(op.jax_expr):
        warn("wrapping operator that has a callable `.jax_expr`")

    def nifty_call(x):
        # Minimal parts that must run outside of JAX
        x = makeField(op.domain, x)
        return op(x).val

    # TODO: define custom JVP and VJP rules
    pt = shapewithdtype_from_domain(op.target, target_dtype)
    hcb_call = partial(pure_callback, nifty_call, pt)

    def wrapped_call(x) -> jft.Vector:
        x = x.tree if isinstance(x, jft.Vector) else x
        return jft.Vector(hcb_call(x))

    return wrapped_call


def convert(
    nifty_obj: Union[Operator, DomainTuple, MultiDomain],
    dtype=float
) -> Union[jft.Model, jft.Vector, jft.ShapeWithDtype, Dict[str,
                                                          jft.ShapeWithDtype]]:
    if not isinstance(nifty_obj, (Operator, DomainTuple, MultiDomain)):
        raise TypeError(f"invalid input type {type(nifty_obj)!r}")

    if isinstance(nifty_obj, (Field, MultiField)):
        return jft.Vector(nifty_obj.val)
    elif isinstance(nifty_obj, (DomainTuple, MultiDomain)):
        return shapewithdtype_from_domain(nifty_obj, dtype)
    else:
        expr = nifty_obj.jax_expr
        parameter_tree = shapewithdtype_from_domain(nifty_obj.domain, dtype)
        if not callable(expr):
            # TODO: implement conversion via host_callback and custom_vjp
            raise NotImplementedError("Sorry, not yet done :(")

    return jft.Model(expr, domain=parameter_tree)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from mpi4py import MPI
from numpy.testing import assert_equal, assert_raises

from ..common import setup_function, teardown_function

comm = MPI.COMM_WORLD
ntask = comm.Get_size()
rank = comm.Get_rank()
master = (rank == 0)
mpi = ntask > 1

if not mpi:
    comm = None

pmp = pytest.mark.parametrize
pms = pytest.mark.skipif


@pms(ntask != 2, reason="requires exactly two mpi tasks")
@pmp('constants', ([], ['a'], ['b'], ['a', 'b']))
@pmp('point_estimates', ([], ['a'], ['b'], ['a', 'b']))
@pmp('mirror_samples', (False, True))
@pmp('mf', (False, True))
@pmp('geo', (False, True))
@pmp('nsamps', (1, 2))
def test_kl(constants, point_estimates, mirror_samples, mf, geo, nsamps):
    if not mf and (len(point_estimates) != 0 or len(constants) != 0):
        return
    dom = ift.RGSpace((12,), (2.12))
    op = ift.HarmonicSmoothingOperator(dom, 3)
    if mf:
        op = ift.ducktape(dom, None, 'a')*(op.ducktape('b'))
    lh = ift.GaussianEnergy(domain=op.target, sampling_dtype=np.float64) @ op
    ic = ift.GradientNormController(iteration_limit=5)
    ic2 = ift.GradientNormController(iteration_limit=5)
    h = ift.StandardHamiltonian(lh, ic_samp=ic)
    mean0 = ift.from_random(h.domain, 'normal')
    args = {'constants': constants,
            'point_estimates': point_estimates,
            'mirror_samples': mirror_samples,
            'n_samples': nsamps,
            'position': mean0,
            'hamiltonian': h,
            'minimizer_sampling': ift.NewtonCG(ic2) if geo else None}
    if isinstance(mean0, ift.MultiField) and set(point_estimates) == set(mean0.keys()):
        with assert_raises(RuntimeError):
            ift.SampledKLEnergy(**args, comm=comm)
        return

    kl0 = ift.SampledKLEnergy(**args, comm=comm)
    if isinstance(mean0, ift.MultiField):
        invariant = list(set(constants).intersection(point_estimates))
        _, tmph = h.simplify_for_constant_input(mean0.extract_by_keys(invariant))
        tmpmean = mean0.extract(tmph.domain)
        invariant = mean0.extract_by_keys(invariant)
    else:
        tmph = h
        tmpmean = mean0
        invariant = None
    samp = kl0._sample_list
    ift.extra.assert_allclose(tmpmean, samp._m)

    if not mpi:
        samples = tuple(s for s in samp._r)
        ii = len(samples)//2
        slc = slice(None, ii) if rank == 0 else slice(ii, None)
        locsamp = samples[slc]
        if mirror_samples:
            neg = [False, ]*2*nsamps if geo else [False, True]*nsamps
        else:
            neg = [False, ]*nsamps
        locneg = neg[slc]
        samp = ift.minimization.sample_list.ResidualSampleList(
                    tmpmean, locsamp, locneg, comm)

    from nifty8.minimization.kl_energies import SampledKLEnergyClass
    kl1 = SampledKLEnergyClass(samp, tmph, constants, invariant, False)

    # Test number of samples
    expected_nsamps = 2*nsamps if mirror_samples else nsamps
    ift.myassert(kl0.samples.n_samples == expected_nsamps)
    ift.myassert(kl1.samples.n_samples == expected_nsamps)

    # Test value
    assert_equal(kl0.value, kl1.value)

    # Test gradient
    if mf:
        for kk in kl0.gradient.domain.keys():
            res0 = kl0.gradient[kk].val
            res1 = kl1.gradient[kk].val
            assert_equal(res0, res1)
    else:
        assert_equal(kl0.gradient.val, kl1.gradient.val)


@pmp('seed', (42, 123))
@pmp('n_samples', (1, 2, 5, 6))
@pmp('geo', (False, True))
def test_mirror(n_samples, seed, geo):
    ift.random.push_sseq_from_seed(seed)
    a = ift.FieldAdapter(ift.UnstructuredDomain(2), 'a').exp()
    lh = ift.GaussianEnergy(domain=a.target, sampling_dtype=float) @ a
    H = ift.StandardHamiltonian(lh, ic_samp=ift.AbsDeltaEnergyController(1E-10, iteration_limit=2))
    mini = None
    if geo:
        mini = ift.NewtonCG(ift.AbsDeltaEnergyController(1E-10, iteration_limit=0))
    KL = ift.SampledKLEnergy(ift.from_random(H.domain), H, n_samples, mini,
                             mirror_samples=True, comm=comm)
    sams = list([s-KL.position for s in KL.samples.iterator()])
    for i in range(len(sams)//2):
        ift.extra.assert_allclose(sams[2*i], -sams[2*i+1])



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from itertools import chain, product

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal, assert_raises

from ..common import setup_function, teardown_function

pmp = pytest.mark.parametrize

HARMONIC_SPACES = [
    ift.RGSpace((8,), harmonic=True),
    ift.RGSpace((7, 8), harmonic=True),
    ift.RGSpace((6, 6), harmonic=True),
    ift.RGSpace((5, 5), harmonic=True),
    ift.RGSpace((4, 5, 7), harmonic=True),
    ift.LMSpace(6),
    ift.LMSpace(9)
]

# Try all sensible kinds of combinations of spaces and binning parameters
CONSISTENCY_CONFIGS_IMPLICIT = product(HARMONIC_SPACES, [None], [None, 3, 4],
                                       [True, False])
CONSISTENCY_CONFIGS_EXPLICIT = product(HARMONIC_SPACES, [[0., 1.3]], [None],
                                       [None])
CONSISTENCY_CONFIGS = chain(CONSISTENCY_CONFIGS_IMPLICIT,
                            CONSISTENCY_CONFIGS_EXPLICIT)

# [harmonic_partner, logarithmic, nbin, binbounds, expected]
CONSTRUCTOR_CONFIGS = [
    [1, False, None, None, {
        'error': (ValueError, NotImplementedError)
    }],
    [ift.RGSpace((8,)), False, None, None, {
        'error': ValueError
    }],
    [
        ift.RGSpace((8,), harmonic=True), None, None, None, {
            'harmonic':
            False,
            'shape': (5,),
            'size':
            5,
            'harmonic_partner':
            ift.RGSpace((8,), harmonic=True),
            'binbounds':
            None,
            'pindex':
            np.array([0, 1, 2, 3, 4, 3, 2, 1]),
            'k_lengths':
            np.array([0., 1., 2., 3., 4.]),
        }
    ],
    [
        ift.RGSpace((8,), harmonic=True), True, None, None, {
            'harmonic':
            False,
            'shape': (4,),
            'size':
            4,
            'harmonic_partner':
            ift.RGSpace((8,), harmonic=True),
            'binbounds': (0.5, 1.3228756555322954, 3.5),
            'pindex':
            np.array([0, 1, 2, 2, 3, 2, 2, 1]),
            'k_lengths':
            np.array([0., 1., 2.5, 4.]),
        }
    ],
]


def k_lengths_configs():
    da_0 = np.array([0, 1.0, 1.41421356, 2., 2.23606798, 2.82842712])
    return [
        [ift.RGSpace((4, 4), harmonic=True), da_0],
    ]


@pmp('attribute, expected_type', [
    ['harmonic_partner', ift.StructuredDomain],
    ['binbounds', type(None)],
    ['pindex', np.ndarray],
    ['k_lengths', np.ndarray],
])
def test_property_ret_type(attribute, expected_type):
    r = ift.RGSpace((4, 4), harmonic=True)
    p = ift.PowerSpace(r)
    ift.myassert(isinstance(getattr(p, attribute), expected_type))


@pmp('harmonic_partner, binbounds, nbin, logarithmic', CONSISTENCY_CONFIGS)
def test_rhopindexConsistency(harmonic_partner, binbounds, nbin, logarithmic):
    bb = ift.PowerSpace.useful_binbounds(harmonic_partner, logarithmic, nbin)
    p = ift.PowerSpace(harmonic_partner=harmonic_partner, binbounds=bb)

    assert_equal(
        np.bincount(p.pindex.ravel()),
        p.dvol,
        err_msg='rho is not equal to pindex degeneracy')


@pmp('harmonic_partner, logarithmic, nbin, binbounds, expected',
     CONSTRUCTOR_CONFIGS)
def test_constructor(harmonic_partner, logarithmic, nbin, binbounds, expected):
    if 'error' in expected:
        with assert_raises(expected['error']):
            bb = ift.PowerSpace.useful_binbounds(harmonic_partner, logarithmic,
                                                 nbin)
            ift.PowerSpace(harmonic_partner=harmonic_partner, binbounds=bb)
    else:
        bb = ift.PowerSpace.useful_binbounds(harmonic_partner, logarithmic,
                                             nbin)
        p = ift.PowerSpace(harmonic_partner=harmonic_partner, binbounds=bb)
        for key, value in expected.items():
            if isinstance(value, np.ndarray):
                assert_allclose(getattr(p, key), value)
            else:
                assert_equal(getattr(p, key), value)


@pmp('harmonic_partner, expected', k_lengths_configs())
def test_k_lengths(harmonic_partner, expected):
    p = ift.PowerSpace(harmonic_partner=harmonic_partner)
    assert_allclose(p.k_lengths, expected)


def test_dvol():
    hp = ift.RGSpace(10, harmonic=True)
    p = ift.PowerSpace(harmonic_partner=hp)
    v1 = hp.dvol
    v1 = hp.size*v1 if np.isscalar(v1) else np.sum(v1)
    v2 = p.dvol
    v2 = p.size*v2 if np.isscalar(v2) else np.sum(v2)
    assert_allclose(v1, v2)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2020 Max-Planck-Society
# Author: Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import pytest

from ..common import setup_function, teardown_function

pmp = pytest.mark.parametrize

dom = ift.RGSpace(5)

@pmp("inp", [(4., True), (-3., False), (2, True), (2.+0j, False), (1, True)])
@pmp("mode", [0, 1, 2])
def test_get_sqrt(inp, mode):
    val, shouldwork = inp
    fld = ift.full(dom, val)
    if mode == 0:
        op = ift.makeOp(fld)
    elif mode == 1:
        op = ift.makeOp(val, ift.makeDomain(dom))
    elif mode == 2:
        fft = ift.FFTOperator(dom)
        if val == 1:
            op = ift.SandwichOperator.make(fft)
        else:
            cheese = ift.makeOp(ift.full(fft.target, val))
            op = ift.SandwichOperator.make(fft, cheese=cheese)
    if not shouldwork:
        with pytest.raises(ValueError):
            op.get_sqrt()
        return
    inp = ift.from_random(op.domain)
    res0 = (op.get_sqrt().adjoint @ op.get_sqrt())(inp)
    res1 = op(inp)
    ift.extra.assert_allclose(res0, res1, rtol=1e-15)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest

from ..common import list2fixture, setup_function, teardown_function

_h_RG_spaces = [
    ift.RGSpace(7, distances=0.2, harmonic=True),
    ift.RGSpace((12, 46), distances=(.2, .3), harmonic=True)
]
_h_spaces = _h_RG_spaces + [ift.LMSpace(17)]
_p_RG_spaces = [
    ift.RGSpace(19, distances=0.7),
    ift.RGSpace((1, 2, 3, 6), distances=(0.2, 0.25, 0.34, .8))
]
_p_spaces = _p_RG_spaces + [ift.HPSpace(17), ift.GLSpace(8, 13)]
_pow_spaces = [ift.PowerSpace(ift.RGSpace((17, 38), harmonic=True))]

pmp = pytest.mark.parametrize
dtype = list2fixture([np.float64, np.complex128])


def _check_repr(op):
    op.__repr__()


@pmp('sp', _p_RG_spaces)
def testLOSResponse(sp, dtype):
    starts = ift.random.current_rng().standard_normal((len(sp.shape), 10))
    ends = ift.random.current_rng().standard_normal((len(sp.shape), 10))
    sigma_low = 1e-4*ift.random.current_rng().standard_normal(10)
    sigma_ups = 1e-5*ift.random.current_rng().standard_normal(10)
    _check_repr(ift.LOSResponse(sp, starts, ends, sigma_low, sigma_ups))


@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
def testOperatorCombinations(sp, dtype):
    a = ift.DiagonalOperator(ift.Field.from_random(sp, "normal", dtype=dtype))
    b = ift.DiagonalOperator(ift.Field.from_random(sp, "normal", dtype=dtype))
    _check_repr(ift.SandwichOperator.make(a, b))
    _check_repr(a(b))
    _check_repr(a+b)
    _check_repr(a-b)
    _check_repr(a*b)
    _check_repr(a**2)


def testLinearInterpolator():
    sp = ift.RGSpace((10, 8), distances=(0.1, 3.5))
    pos = ift.random.current_rng().random((2, 23))
    pos[0, :] *= 0.9
    pos[1, :] *= 7*3.5
    _check_repr(ift.LinearInterpolator(sp, pos))


@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
def testOperatorAdaptor(sp, dtype):
    op = ift.DiagonalOperator(ift.Field.from_random(sp, "normal", dtype=dtype))
    _check_repr(op.adjoint)
    _check_repr(op.inverse)
    _check_repr(op.inverse.adjoint)


@pmp('sp1', _h_spaces + _p_spaces + _pow_spaces)
@pmp('sp2', _h_spaces + _p_spaces + _pow_spaces)
def testNullOperator(sp1, sp2, dtype):
    op = ift.NullOperator(sp1, sp2)
    ift.extra.check_linear_operator(op, dtype, dtype)
    mdom1 = ift.MultiDomain.make({'a': sp1})
    mdom2 = ift.MultiDomain.make({'b': sp2})
    _check_repr(ift.NullOperator(mdom1, mdom2))
    _check_repr(ift.NullOperator(sp1, mdom2))
    _check_repr(ift.NullOperator(mdom1, sp2))


@pmp('sp', _p_RG_spaces)
def testHarmonicSmoothingOperator(sp, dtype):
    _check_repr(ift.HarmonicSmoothingOperator(sp, 0.1))


@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
def testDOFDistributor(sp, dtype):
    # TODO: Test for DomainTuple
    if sp.size < 4:
        return
    dofdex = np.arange(sp.size).reshape(sp.shape) % 3
    dofdex = ift.Field.from_raw(sp, dofdex)
    _check_repr(ift.DOFDistributor(dofdex))


@pmp('sp', _h_spaces)
def testPPO(sp, dtype):
    _check_repr(ift.PowerDistributor(target=sp))
    ps = ift.PowerSpace(
        sp, ift.PowerSpace.useful_binbounds(sp, logarithmic=False, nbin=3))
    _check_repr(ift.PowerDistributor(target=sp, power_space=ps))
    ps = ift.PowerSpace(
        sp, ift.PowerSpace.useful_binbounds(sp, logarithmic=True, nbin=3))
    _check_repr(ift.PowerDistributor(target=sp, power_space=ps))


@pmp('sp', _h_RG_spaces + _p_RG_spaces)
def testFFT(sp, dtype):
    _check_repr(ift.FFTOperator(sp))
    _check_repr(ift.FFTOperator(sp.get_default_codomain()))


@pmp('sp', _h_RG_spaces + _p_RG_spaces)
def testHartley(sp, dtype):
    _check_repr(ift.HartleyOperator(sp))
    _check_repr(ift.HartleyOperator(sp.get_default_codomain()))


@pmp('sp', _h_spaces)
def testHarmonic(sp, dtype):
    _check_repr(ift.HarmonicTransformOperator(sp))


@pmp('sp', _p_spaces)
def testMask(sp, dtype):
    # Create mask
    f = ift.from_random(sp, 'normal').val
    mask = np.zeros_like(f)
    mask[f > 0] = 1
    mask = ift.Field.from_raw(sp, mask)
    # Test MaskOperator
    _check_repr(ift.MaskOperator(mask))


@pmp('sp', _h_spaces + _p_spaces)
def testDiagonal(sp, dtype):
    op = ift.DiagonalOperator(ift.Field.from_random(sp, "normal", dtype=dtype))
    _check_repr(op)


@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
def testGeometryRemover(sp, dtype):
    _check_repr(ift.GeometryRemover(sp))


@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
def testExtractAtIndices(sp, dtype):
    min_ax = np.min(sp.shape)
    n_ax = len(sp.shape)
    inds = (list(range(min_ax//2))+list(range(min_ax//3)), )*n_ax
    _check_repr(ift.ExtractAtIndices(sp, inds))


@pmp('spaces', [0, 1, 2, 3, (0, 1), (0, 2), (0, 1, 2), (0, 2, 3), (1, 3)])
@pmp('wgt', [0, 1, 2, -1])
def testContractionOperator(spaces, wgt, dtype):
    dom = (ift.RGSpace(10), ift.RGSpace(13), ift.GLSpace(5), ift.HPSpace(4))
    _check_repr(ift.ContractionOperator(dom, spaces, wgt))


def testDomainTupleFieldInserter():
    target = ift.DomainTuple.make((ift.UnstructuredDomain([3, 2]),
                                   ift.UnstructuredDomain(7),
                                   ift.RGSpace([4, 22])))
    _check_repr(ift.DomainTupleFieldInserter(target, 1, (5,)))


@pmp('space', [0, 2])
@pmp('factor', [1, 2, 2.7])
@pmp('central', [False, True])
def testZeroPadder(space, factor, dtype, central):
    dom = (ift.RGSpace(10), ift.UnstructuredDomain(13), ift.RGSpace(7, 12),
           ift.HPSpace(4))
    newshape = [int(factor*l) for l in dom[space].shape]
    _check_repr(ift.FieldZeroPadder(dom, newshape, space, central))


@pmp('args', [[ift.RGSpace(
    (13, 52, 40)), (4, 6, 25), None], [ift.RGSpace(
        (128, 128)), (45, 48), 0], [ift.RGSpace(13), (7,), None], [
            (ift.HPSpace(3), ift.RGSpace((12, 24), distances=0.3)), (12, 12), 1
        ]])
def testRegridding(args):
    _check_repr(ift.RegriddingOperator(*args))


@pmp('fdomain', [
    ift.DomainTuple.make((ift.RGSpace(
        (3, 5, 4)), ift.RGSpace((16,), distances=(7.,))),),
    ift.DomainTuple.make(ift.HPSpace(12),)
])
@pmp('domain', [
    ift.DomainTuple.make((ift.RGSpace((2,)), ift.GLSpace(10)),),
    ift.DomainTuple.make(ift.RGSpace((10, 12), distances=(0.1, 1.)),)
])
def testOuter(fdomain, domain):
    f = ift.from_random(fdomain, 'normal')
    _check_repr(ift.OuterProduct(domain, f))


@pmp('sp', _h_spaces + _p_spaces + _pow_spaces)
@pmp('seed', [12, 3])
def testValueInserter(sp, seed):
    with ift.random.Context(seed):
        ind = []
        for ss in sp.shape:
            if ss == 1:
                ind.append(0)
            else:
                ind.append(int(ift.random.current_rng().integers(0, ss - 1)))
        _check_repr(ift.ValueInserter(sp, ind))


@pmp('dom', _h_spaces + _p_spaces)
def testGaussianEnergy(dom):
    _check_repr(ift.GaussianEnergy(domain=dom))


@pmp('dom', _h_spaces + _p_spaces)
def testHamiltonian(dom):
    lh = ift.GaussianEnergy(domain=dom)
    _check_repr(ift.StandardHamiltonian(lh))


@pmp('dom', _h_RG_spaces + _p_RG_spaces)
def testSumOperator(dom):
    op0 = ift.FFTOperator(dom)
    op1 = op0.scale(2)
    _check_repr(op0+op1)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2022 Max-Planck-Society

import nifty8 as ift
import numpy as np
import pytest

from ..common import list2fixture, setup_function, teardown_function

pmp = pytest.mark.parametrize

dom0 = ift.RGSpace([2, 43])
dom1 = ift.UnstructuredDomain([43, 1, 1])
dom2 = ift.HPSpace(3)


def test_transpose_operator_trivial():
    op = ift.ScalingOperator(dom0, 1.).transpose((0,))
    ift.extra.check_linear_operator(op)
    fld = ift.from_random(op.domain)
    ift.extra.assert_equal(op(fld), fld)


index_pairs = list2fixture([
               [(0, 1, 2), (0, 1, 2, 3, 4, 5)],
               [(0, 2, 1), (0, 1, 5, 2, 3, 4)],
               [(1, 0, 2), (2, 3, 4, 0, 1, 5)],
               [(1, 2, 0), (2, 3, 4, 5, 0, 1)],
               [(2, 1, 0), (5, 2, 3, 4, 0, 1)],
               [(2, 0, 1), (5, 0, 1, 2, 3, 4)],
              ])


def test_transpose_operator(index_pairs):
    dom = ift.makeDomain((dom0, dom1, dom2))
    space_indices, np_indices = index_pairs

    np_indices1 = ift.operators.transpose_operator._niftyspace_to_np_indices(dom, space_indices)
    assert np_indices == np_indices1

    op = ift.ScalingOperator(dom, 1.).transpose(space_indices)
    ift.extra.check_linear_operator(op)
    fld = ift.from_random(op.domain)
    res0 = op(fld)
    res1 = ift.makeField(op.target, np.transpose(fld.val, np_indices))
    ift.extra.assert_equal(res0, res1)


def test_transpose_operator_fail():
    iden = ift.ScalingOperator(dom0, 1.)
    with pytest.raises(IndexError):
        iden.transpose((1,))
    with pytest.raises(IndexError):
        iden.transpose((0, 1))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from . import utilities
from .domain_tuple import DomainTuple
from .field import Field
from .multi_domain import MultiDomain
from .operators.operator import Operator


class MultiField(Operator):
    def __init__(self, domain, val):
        """The discrete representation of a continuous field over a sum space.

        Parameters
        ----------
        domain: MultiDomain
        val: tuple containing Field entries
        """
        if not isinstance(domain, MultiDomain):
            raise TypeError("domain must be of type MultiDomain")
        if not isinstance(val, tuple):
            raise TypeError("val must be a tuple")
        if len(val) != len(domain):
            raise ValueError("length mismatch")
        for d, v in zip(domain._domains, val):
            if isinstance(v, Field):
                utilities.check_object_identity(v._domain, d)
            else:
                raise TypeError("bad entry in val (must be Field)")
        self._domain = domain
        self._val = val

    @staticmethod
    def from_dict(dct, domain=None):
        if domain is None:
            for dd in dct.values():
                if not isinstance(dd.domain, DomainTuple):
                    raise TypeError('Values of dictionary need to be Fields '
                                    'defined on DomainTuples.')
            domain = MultiDomain.make({key: v._domain
                                       for key, v in dct.items()})
        res = tuple(dct[key] if key in dct else Field(dom, 0.)
                    for key, dom in zip(domain.keys(), domain.domains()))
        return MultiField(domain, res)

    def to_dict(self):
        return {key: val for key, val in zip(self._domain.keys(), self._val)}

    def __getitem__(self, key):
        return self._val[self._domain.idx[key]]

    def __contains__(self, key):
        return key in self._domain.idx

    def keys(self):
        return self._domain.keys()

    def items(self):
        return zip(self._domain.keys(), self._val)

    def values(self):
        return self._val

    @property
    def domain(self):
        return self._domain

    @property
    def dtype(self):
        return {key: val.dtype for key, val in self.items()}

    def _transform(self, op):
        return MultiField(self._domain, tuple(op(v) for v in self._val))

    @property
    def real(self):
        """MultiField : The real part of the multi field"""
        return self._transform(lambda x: x.real)

    @property
    def imag(self):
        """MultiField : The imaginary part of the multi field"""
        return self._transform(lambda x: x.imag)

    @staticmethod
    def from_random(domain, random_type='normal', dtype=np.float64, **kwargs):
        """Draws a random multi-field with the given parameters.

        Parameters
        ----------
        random_type : 'pm1', 'normal', or 'uniform'
            The random distribution to use.
        domain : DomainTuple
            The domain of the output random Field.
        dtype : type
            The datatype of the output random Field.
            If the datatype is complex, each real an imaginary part have
            variance 1.

        Returns
        -------
        MultiField
            The newly created :class:`MultiField`.

        Notes
        -----
        The individual fields within this multi-field will be drawn in alphabetical
        order of the multi-field's domain keys. As a consequence, renaming these
        keys may cause the multi-field to be filled with different random numbers,
        even for the same initial RNG state.
        """
        domain = MultiDomain.make(domain)
        if isinstance(dtype, dict):
            dtype = {kk: np.dtype(dt) for kk, dt in dtype.items()}
        else:
            dtype = np.dtype(dtype)
            dtype = {kk: dtype for kk in domain.keys()}
        dct = {kk: Field.from_random(domain[kk], random_type, dtype[kk], **kwargs)
               for kk in domain.keys()}
        return MultiField.from_dict(dct)

    def s_vdot(self, x):
        result = 0.
        utilities.check_object_identity(x._domain, self._domain)
        for v1, v2 in zip(self._val, x._val):
            result += v1.s_vdot(v2)
        return result

    def vdot(self, x):
        return Field.scalar(self.s_vdot(x))

#    @staticmethod
#    def build_dtype(dtype, domain):
#        if isinstance(dtype, dict):
#            return dtype
#        if dtype is None:
#            dtype = np.float64
#        return {key: dtype for key in domain.keys()}

    @staticmethod
    def full(domain, val):
        domain = MultiDomain.make(domain)
        return MultiField(domain, tuple(Field(dom, val)
                          for dom in domain._domains))

    @property
    def val(self):
        return {key: val.val
                for key, val in zip(self._domain.keys(), self._val)}

    def val_rw(self):
        return {key: val.val_rw()
                for key, val in zip(self._domain.keys(), self._val)}

    @staticmethod
    def from_raw(domain, arr):
        return MultiField(
            domain, tuple(Field(domain[key], arr[key])
                          for key in domain.keys()))

    def norm(self, ord=2):
        """Computes the norm of the field values.

        Parameters
        ----------
        ord : int, default=2
            accepted values: 1, 2, ..., np.inf

        Returns
        -------
        norm : float
            The norm of the field values.
        """
        nrm = np.asarray([f.norm(ord) for f in self._val])
        if ord == np.inf:
            return nrm.max()
        return (nrm ** ord).sum() ** (1./ord)
#        return np.sqrt(np.abs(self.vdot(x=self)))

    def s_sum(self):
        """Computes the sum all field values.

        Returns
        -------
        norm : float
            The sum of the field values.
        """
        return utilities.my_sum(map(lambda v: v.s_sum(), self._val))

    @property
    def size(self):
        """Computes the overall degrees of freedom.

        Returns
        -------
        size : int
            The sum of the size of the individual fields
        """
        return utilities.my_sum(map(lambda d: d.size, self._domain.domains()))

    def __neg__(self):
        return self._transform(lambda x: -x)

    def __abs__(self):
        return self._transform(lambda x: abs(x))

    def conjugate(self):
        return self._transform(lambda x: x.conjugate())

    def clip(self, a_min=None, a_max=None):
        return self.ptw("clip", a_min, a_max)

    def s_all(self):
        for v in self._val:
            if not v.s_all():
                return False
        return True

    def s_any(self):
        for v in self._val:
            if v.s_any():
                return True
        return False

    def extract(self, subset):
        if subset is self._domain:
            return self
        return MultiField(subset,
                          tuple(self[key] for key in subset.keys()))

    def extract_by_keys(self, keys):
        dom = MultiDomain.make({kk: vv for kk, vv in self.domain.items() if kk in keys})
        return self.extract(dom)

    def extract_part(self, subset):
        if subset is self._domain:
            return self
        dct = {key: self[key] for key in subset.keys() if key in self}
        if len(dct) == 0:
            return None
        return MultiField.from_dict(dct)

    def unite(self, other):
        """Merges two MultiFields on potentially different MultiDomains.

        Parameters
        ----------
        other : :class:`nifty8.multi_field.MultiField`
            the partner Field

        Returns
        -------
        MultiField
            This MultiField's domain is the union of the input fields'
            domains. The values are the sum of the fields in self and other.
            If a field is not present, it is assumed to have an uniform value
            of zero.
        """
        return self.flexible_addsub(other, False)

    @staticmethod
    def union(fields, domain=None):
        """Returns the union of its input fields.

        Parameters
        ----------
        fields : iterable of :class:`nifty8.multi_field.MultiField`
            The set of input fields. Their domains need not be identical.
        domain : MultiDomain or None
            If supplied, this will be the domain of the resulting field.
            Providing this domain will accelerate the function.

        Returns
        -------
        MultiField
            The union of the input fields

        Notes
        -----
        If the same key occurs more than once in the input fields, the value
        associated with the last occurrence will be put into the output.
        No summation is performed!
        """
        res = {}
        for field in fields:
            res.update(field.to_dict())
        return MultiField.from_dict(res, domain)

    def flexible_addsub(self, other, neg):
        """Merges two MultiFields on potentially different MultiDomains.

        Parameters
        ----------
        other : :class:`nifty8.multi_field.MultiField`
            the partner Field
        neg : bool or dict
            if True, the partner field is subtracted, otherwise added

        Returns
        -------
        MultiField
            This MultiField's domain is the union of the input fields'
            domains. The values are the sum (or difference, if neg==True) of
            the fields in self and other. If a field is not present, it is
            assumed to have an uniform value of zero.
        """
        if self._domain is other._domain and not isinstance(neg, dict):
            return self-other if neg else self+other

        if isinstance(neg, dict):
            fct = lambda k: neg[k]
        else:
            fct = lambda k: neg
        res = self.to_dict()
        for key, val in other.items():
            if key in res:
                res[key] = res[key]-val if fct(key) else res[key]+val
            else:
                res[key] = -val if fct(key) else val
        return MultiField.from_dict(res)

    def _prep_args(self, args, kwargs, i):
        for arg in args + tuple(kwargs.values()):
            if not (arg is None or np.isscalar(arg) or arg.jac is None):
                raise TypeError("bad argument")
        argstmp = tuple(arg if arg is None or np.isscalar(arg) else arg._val[i]
                        for arg in args)
        kwargstmp = {key: val if val is None or np.isscalar(val) else val._val[i]
                     for key, val in kwargs.items()}
        return argstmp, kwargstmp

    def ptw(self, op, *args, **kwargs):
        tmp = []
        for i in range(len(self._val)):
            argstmp, kwargstmp = self._prep_args(args, kwargs, i)
            tmp.append(self._val[i].ptw(op, *argstmp, **kwargstmp))
        return MultiField(self.domain, tuple(tmp))

    def ptw_with_deriv(self, op, *args, **kwargs):
        tmp = []
        for i in range(len(self._val)):
            argstmp, kwargstmp = self._prep_args(args, kwargs, i)
            tmp.append(self._val[i].ptw_with_deriv(op, *argstmp, **kwargstmp))
        return (MultiField(self.domain, tuple(v[0] for v in tmp)),
                MultiField(self.domain, tuple(v[1] for v in tmp)))

    def _binary_op(self, other, op):
        f = getattr(Field, op)
        if isinstance(other, MultiField):
            utilities.check_object_identity(self._domain, other._domain)
            val = tuple(f(v1, v2)
                        for v1, v2 in zip(self._val, other._val))
        else:
            val = tuple(f(v1, other) for v1 in self._val)
        return MultiField(self._domain, val)


for op in ["__add__", "__radd__",
           "__sub__", "__rsub__",
           "__mul__", "__rmul__",
           "__truediv__", "__rtruediv__",
           "__floordiv__", "__rfloordiv__",
           "__pow__", "__rpow__",
           "__lt__", "__le__", "__gt__", "__ge__", "__eq__", "__ne__"]:
    def func(op):
        def func2(self, other):
            return self._binary_op(other, op)
        return func2
    setattr(MultiField, op, func(op))


for op in ["__iadd__", "__isub__", "__imul__", "__idiv__",
           "__itruediv__", "__ifloordiv__", "__ipow__"]:
    def func(op):
        def func2(self, other):
            raise TypeError(
                "In-place operations are deliberately not supported")
        return func2
    setattr(MultiField, op, func(op))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2021 Max-Planck-Society
# Author: Philipp Arras

import nifty8 as ift
import numpy as np
import pytest

try:
    import jax.numpy as jnp
except ImportError:
    pass

from ..common import setup_function, teardown_function

pmp = pytest.mark.parametrize


@pmp("dom", [ift.RGSpace((10, 8)), (ift.RGSpace(10), ift.RGSpace(8))])
@pmp("func", [(lambda x: x, True), (lambda x: x**2, False), (lambda x: x*x, False),
              (lambda x: x*x[0, 0], False), (lambda x: x+x[0, 0], True),
              (lambda x: jnp.sin(x), False), (lambda x: x*x.sum(), False),
              (lambda x: x+x.sum(), True)])
def test_jax(dom, func):
    pytest.importorskip("jax")
    loc = ift.from_random(dom)
    f, linear = func
    res0 = np.array(f(loc.val))
    op = ift.JaxOperator(dom, dom, f)
    np.testing.assert_allclose(res0, op(loc).val)
    ift.extra.check_operator(op, ift.from_random(op.domain))

    op = ift.JaxLinearOperator(dom, dom, f, np.float64)
    if linear:
        ift.extra.check_linear_operator(op)
    else:
        with pytest.raises(Exception):
            ift.extra.check_linear_operator(op)


def test_mf_jax():
    pytest.importorskip("jax")
    dom = ift.makeDomain({"a": ift.RGSpace(10), "b": ift.UnstructuredDomain(2)})

    func = lambda x: x["a"]*x["b"][0]
    op = ift.JaxOperator(dom, dom["a"], func)
    loc = ift.from_random(op.domain)
    np.testing.assert_allclose(np.array(func(loc.val)), op(loc).val)
    ift.extra.check_operator(op, loc)

    func = lambda x: {"a": jnp.full(dom["a"].shape, 2.)*x[0]*x[1], "b": jnp.full(dom["b"].shape, 1.)*jnp.exp(x[0])}
    op = ift.JaxOperator(dom["b"], dom, func)
    loc = ift.from_random(op.domain)
    for kk in dom.keys():
        np.testing.assert_allclose(np.array(func(loc.val)[kk]), op(loc)[kk].val)
    ift.extra.check_operator(op, loc)


@pmp("dom", [ift.RGSpace((10, 8)),
    {"a": ift.RGSpace(10), "b": ift.UnstructuredDomain(2)}])
def test_jax_energy(dom):
    pytest.importorskip("jax")
    dom = ift.makeDomain(dom)
    e0 = ift.GaussianEnergy(domain=dom, sampling_dtype=np.float64)
    def func(x):
        return 0.5*jnp.vdot(x, x)
    def funcmf(x):
        res = 0
        for kk, vv in x.items():
            res += jnp.vdot(vv, vv)
        return 0.5*res
    e = ift.JaxLikelihoodEnergyOperator(dom,
            funcmf if isinstance(dom, ift.MultiDomain) else func,
            transformation=ift.ScalingOperator(dom, 1.),
            sampling_dtype=np.float64)
    ift.extra.check_operator(e, ift.from_random(e.domain))
    for wm in [False, True]:
        pos = ift.from_random(e.domain)
        lin = ift.Linearization.make_var(pos, wm)
        ift.extra.assert_allclose(e0(pos), e(pos))
        ift.extra.assert_allclose(e0(lin).val, e(lin).val)
        ift.extra.assert_allclose(e0(lin).gradient, e(lin).gradient)
        if not wm:
            continue
        pos1 = ift.from_random(e.domain)
        ift.extra.assert_allclose(e0(lin).metric(pos1), e(lin).metric(pos1))


def test_jax_errors():
    pytest.importorskip("jax")
    dom = ift.UnstructuredDomain(2)
    mdom = {"a": dom}
    op = ift.JaxOperator(dom, dom, lambda x: {"a": x})
    fld = ift.full(dom, 0.)
    with pytest.raises(TypeError):
        op(fld)
    op = ift.JaxOperator(dom, mdom, lambda x: x)
    with pytest.raises(TypeError):
        op(fld)
    op = ift.JaxOperator(dom, dom, lambda x: x[0])
    with pytest.raises(ValueError):
        op(fld)
    op = ift.JaxOperator(dom, mdom, lambda x: {"a": x[0]})
    with pytest.raises(ValueError):
        op(fld)


def test_jax_complex():
    pytest.importorskip("jax")
    dom = ift.UnstructuredDomain(1)
    a = ift.ducktape(dom, None, "a")
    b = ift.ducktape(dom, None, "b")
    op = a.real+1j*b.real
    op1 = ift.JaxOperator(op.domain, op.target, lambda x: x["a"] + 1j*x["b"])
    _op_equal(op, op1, ift.from_random(op.domain))
    ift.extra.check_operator(op, ift.from_random(op.domain), ntries=10)
    ift.extra.check_operator(op1, ift.from_random(op.domain), ntries=10)

    op = op.imag
    op1 = op1.imag
    _op_equal(op, op1, ift.from_random(op.domain))
    ift.extra.check_operator(op, ift.from_random(op.domain), ntries=10)
    ift.extra.check_operator(op1, ift.from_random(op.domain), ntries=10)

    lin = ift.Linearization.make_var(ift.from_random(op.domain))
    test_vec = ift.full(op.target, 1.)
    grad = op(lin).jac.adjoint(test_vec)
    grad1 = op1(lin).jac.adjoint(test_vec)
    ift.extra.assert_equal(grad, grad1)
    ift.extra.assert_equal(grad, ift.makeField(grad.domain, {"a": 0., "b": 1.}))


def _op_equal(op0, op1, loc):
    assert op0.domain is op1.domain
    assert op0.target is op1.target
    ift.extra.assert_allclose(op0(loc), op1(loc))
    lin = ift.Linearization.make_var(loc)
    res = op0(lin)
    res1 = op1(lin)
    ift.extra.assert_allclose(res.val, res1.val)
    fld = ift.from_random(op0.domain, dtype=loc.dtype)
    ift.extra.assert_allclose(res.jac(fld), res1.jac(fld))
    fld = ift.from_random(op0.target, dtype=res.val.dtype)
    ift.extra.assert_allclose(res.jac.adjoint(fld), res1.jac.adjoint(fld))



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.7
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# # Custom nonlinearities

# +
import numpy as np

import nifty8 as ift


# -

# In NIFTy, users can add hand-crafted point-wise nonlinearities that are then available for `Field`, `MultiField`, `Linearization` and `Operator`. This guide illustrates how this is done.
#
# Suppose that we would like to use the point-wise function f(x) = x*exp(x) in an operator chain. This function is called "myptw" in the following. We introduce this function to NIFTy by implementing two functions.
#
# First, one that takes a `numpy.ndarray` as an input, applies the point-wise mapping and returns the result as a `numpy.ndarray` of the same shape. Second, a function that takes a `numpy.ndarray` as an input and returns two `numpy.ndarray`s: the application of the nonlinearity (same as before) and the derivative.

# +
def func(x):
    return x*np.exp(x)


def func_and_derv(x):
    expx = np.exp(x)
    return x*expx, (1+x)*expx


# -

# These two functions are then added to the NIFTy-internal dictionary that contains all implemented point-wise nonlinearities.

ift.pointwise.ptw_dict["myptw"] = func, func_and_derv

# This allows us to apply this non-linearity on `Field`s, ...

dom = ift.UnstructuredDomain(10)
fld = ift.from_random(dom)
fld = ift.full(dom, 2.)
a = fld.ptw("myptw")
b = ift.makeField(dom, func(fld.val))
ift.extra.assert_allclose(a, b)

# `MultiField`s, ...

mdom = ift.makeDomain({"bar": ift.UnstructuredDomain(10)})
mfld = ift.from_random(mdom)
a = mfld.ptw("myptw")
b = ift.makeField(mdom, {"bar": func(mfld["bar"].val)})
ift.extra.assert_allclose(a, b)

# `Linearization`s (including the Jacobian), ...

lin = ift.Linearization.make_var(fld)
a = lin.ptw("myptw").val
b = ift.makeField(dom, func(fld.val))
ift.extra.assert_allclose(a, b)

op_a = lin.ptw("myptw").jac
op_b = ift.makeOp(ift.makeField(dom, func_and_derv(fld.val)[1]))
testing_vector = ift.from_random(dom)
ift.extra.assert_allclose(op_a(testing_vector),
                          op_b(testing_vector))

# and `Operator`s.

op = ift.FieldAdapter(dom, "foo").ptw("myptw")


# Please remember to always check that the gradient has been implemented correctly by comparing it to an approximation to the gradient by finite differences.

# +
def check(func_name, eps=1e-7):
    pos = ift.from_random(ift.UnstructuredDomain(10))
    var0 = ift.Linearization.make_var(pos)
    var1 = ift.Linearization.make_var(pos+eps)
    df0 = (var1.ptw(func_name).val - var0.ptw(func_name).val)/eps
    df1 = var0.ptw(func_name).jac(ift.full(lin.domain, 1.))
    # rtol depends on how nonlinear the function is
    ift.extra.assert_allclose(df0, df1, rtol=100*eps)

check("myptw")



#!/usr/bin/env python3

import sys

import pytest

pytest.importorskip("jax")

import jax.numpy as jnp
from jax import random, value_and_grad
from numpy.testing import assert_allclose

import nifty8.re as jft

pmp = pytest.mark.parametrize


def rosenbrock(np):
    def func(x):
        return jnp.sum(100. * jnp.diff(x)**2 + (1. - x[:-1])**2)

    return func


def himmelblau(np):
    def func(p):
        x, y = p
        return (x**2 + y - 11.)**2 + (x + y**2 - 7.)**2

    return func


def matyas(np):
    def func(p):
        x, y = p
        return 0.26 * (x**2 + y**2) - 0.48 * x * y

    return func


def eggholder(np):
    def func(p):
        x, y = p
        return -(y + 47) * jnp.sin(
            jnp.sqrt(jnp.abs(x / 2. + y + 47.))
        ) - x * jnp.sin(jnp.sqrt(jnp.abs(x - (y + 47.))))

    return func


def test_ncg_for_pytree():
    pos = jft.Vector(
        [
            jnp.array(0., dtype=jnp.float32),
            (jnp.array(3., dtype=jnp.float32), ), {
                "a": jnp.array(5., dtype=jnp.float32)
            }
        ]
    )
    getters = (lambda x: x[0], lambda x: x[1][0], lambda x: x[2]["a"])
    tgt = [-10., 1., 2.]
    met = [10., 40., 2]

    def model(p):
        losses = []
        for i, get in enumerate(getters):
            losses.append((get(p) - tgt[i])**2 * met[i])
        return jnp.sum(jnp.array(losses))

    def metric(p, tan):
        m = []
        m.append(tan[0] * met[0])
        m.append((tan[1][0] * met[1], ))
        m.append({"a": tan[2]["a"] * met[2]})
        return jft.Vector(m)

    res = jft.newton_cg(
        fun_and_grad=value_and_grad(model),
        x0=pos,
        hessp=metric,
        maxiter=10,
        absdelta=1e-6
    )
    for i, get in enumerate(getters):
        assert_allclose(get(res), tgt[i], atol=1e-6, rtol=1e-5)


@pmp("seed", (3637, 12, 42))
def test_ncg(seed):
    key = random.PRNGKey(seed)
    x = random.normal(key, shape=(3, ))
    diag = jnp.array([1., 2., 3.])
    met = lambda y, t: t / diag
    val_and_grad = lambda y: (
        jnp.sum(y**2 / diag) / 2 - jnp.dot(x, y), y / diag - x
    )

    res = jft.newton_cg(
        fun_and_grad=val_and_grad,
        x0=x,
        hessp=met,
        maxiter=20,
        absdelta=1e-6,
        name='N'
    )
    assert_allclose(res, diag * x, rtol=1e-4, atol=1e-4)


@pmp("seed", (3637, 12, 42))
@pmp("cg", (jft.cg, jft.static_cg))
def test_cg(seed, cg):
    key = random.PRNGKey(seed)
    sk = random.split(key, 2)
    x = random.normal(sk[0], shape=(3, ))
    # Avoid poorly conditioned matrices by shifting the elements from zero
    diag = 6. + random.normal(sk[1], shape=(3, ))
    mat = lambda x: x / diag

    res, _ = cg(mat, x, resnorm=1e-5, absdelta=1e-5)
    assert_allclose(res, diag * x, rtol=1e-4, atol=1e-4)


@pmp("seed", (3637, 12, 42))
@pmp("cg", (jft.cg, jft.static_cg))
def test_cg_non_pos_def_failure(seed, cg):
    key = random.PRNGKey(seed)
    sk = random.split(key, 2)

    x = random.normal(sk[0], shape=(4, ))
    # Purposely produce a non-positive definite matrix
    diag = jnp.concatenate(
        (jnp.array([-1]), 6. + random.normal(sk[1], shape=(3, )))
    )
    mat = lambda x: x / diag

    with pytest.raises(ValueError):
        _, info = cg(mat, x, resnorm=1e-5, absdelta=1e-5)
        if info < 0:
            raise ValueError()


@pmp("seed", (3637, 12, 42))
def test_cg_steihaug(seed):
    key = random.PRNGKey(seed)
    sk = random.split(key, 2)
    x = random.normal(sk[0], shape=(3, ))
    # Avoid poorly conditioned matrices by shifting the elements from zero
    diag = 6. + random.normal(sk[1], shape=(3, ))
    mat = lambda x: x / diag

    # Note, the solution to the subproblem with infinite trust radius is the CG
    # but with the opposite sign
    res = jft.conjugate_gradient._cg_steihaug_subproblem(
        jnp.nan, -x, mat, resnorm=1e-6, trust_radius=jnp.inf
    )
    assert_allclose(res.step, diag * x, rtol=1e-4, atol=1e-4)


@pmp("seed", (3637, 12, 42))
@pmp("size", (5, 9, 14))
def test_cg_steihaug_vs_cg_consistency(seed, size):
    key = random.PRNGKey(seed)
    sk = random.split(key, 2)

    x = random.normal(sk[0], shape=(size, ))
    # Avoid poorly conditioned matrices by shifting the elements from zero
    mat_val = 6. + random.normal(sk[1], shape=(size, size))
    mat_val = mat_val @ mat_val.T  # Construct a symmetric matrix
    mat = lambda x: mat_val @ x

    # Note, the solution to the subproblem with infinite trust radius is the CG
    # but with the opposite sign
    for i in range(4):
        print(f"Iteratoin {i:02d}", file=sys.stderr)
        res_cgs = jft.conjugate_gradient._cg_steihaug_subproblem(
            jnp.nan,
            -x,
            mat,
            resnorm=1e-6,
            trust_radius=jnp.inf,
            miniter=i,
            maxiter=i
        )
        res_cg_plain, _ = jft.conjugate_gradient.cg(
            mat, x, resnorm=1e-6, miniter=i, maxiter=i
        )
        assert_allclose(res_cgs.step, res_cg_plain, rtol=1e-4, atol=1e-5)


@pmp(
    "fun_and_init",
    (
        (rosenbrock, jnp.zeros(2)),
        (himmelblau, jnp.zeros(2)),
        (matyas, jnp.ones(2) * 6.),
        # (eggholder, jnp.ones(2) * 100.)  # `eggholder` potential leads to
        # infinite loop in trust-ncg's subproblem solver, see
        # https://github.com/google/jax/issues/15035
    )
)
@pmp("maxiter", (jnp.inf, None))
def test_minimize(fun_and_init, maxiter):
    from jax import grad, hessian
    from scipy.optimize import minimize as opt_minimize

    func, x0 = fun_and_init

    def jft_minimize(x0):
        result = jft.minimize(
            func(jnp),
            x0,
            method='trust-ncg',
            options=dict(
                maxiter=maxiter,
                energy_reduction_factor=None,
                gtol=1e-6,
                initial_trust_radius=1.,
                max_trust_radius=1000.
            ),
        )
        return result.x

    def scp_minimize(x0):
        # Use JAX primitives to take derivates
        fun = func(jnp)
        result = opt_minimize(
            fun, x0, jac=grad(fun), hess=hessian(fun), method='trust-ncg'
        )
        return result.x

    jax_res = jft_minimize(x0)
    scipy_res = scp_minimize(x0)
    assert_allclose(scipy_res, jax_res, rtol=2e-6, atol=2e-5)


if __name__ == "__main__":
    test_ncg_for_pytree()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2022 Max-Planck-Society, Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_equal

from ..common import list2fixture, setup_function, teardown_function


def test_counting_operator():
    dom = ift.UnstructuredDomain(2)
    inp = ift.Operator.identity_operator(dom)

    counting = ift.CountingOperator(dom)

    op = inp.exp() @ counting

    assert_equal(counting.count_apply, 0)
    assert_equal(counting.count_apply_lin, 0)
    assert_equal(counting.count_jac, 0)
    assert_equal(counting.count_jac_adj, 0)

    fld = ift.from_random(op.domain)
    lin = ift.Linearization.make_var(fld)
    op(fld)
    op(fld)

    assert_equal(counting.count_apply, 2)
    assert_equal(counting.count_apply_lin, 0)
    assert_equal(counting.count_jac, 0)
    assert_equal(counting.count_jac_adj, 0)

    res = op(lin)

    assert_equal(counting.count_apply, 2)
    assert_equal(counting.count_apply_lin, 1)
    assert_equal(counting.count_jac, 0)
    assert_equal(counting.count_jac_adj, 0)

    for _ in range(3):
        res.jac(fld)

    assert_equal(counting.count_apply, 2)
    assert_equal(counting.count_apply_lin, 1)
    assert_equal(counting.count_jac, 3)
    assert_equal(counting.count_jac_adj, 0)

    for _ in range(5):
        res.jac.adjoint(res.val)

    assert_equal(counting.count_apply, 2)
    assert_equal(counting.count_apply_lin, 1)
    assert_equal(counting.count_jac, 3)
    assert_equal(counting.count_jac_adj, 5)

    counting.__repr__()

    ift.extra.check_operator(counting, fld)



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause
# Authors: Philipp Frank, Jakob Roth, Gordian Edenhofer

import inspect
import os
import pickle
from dataclasses import field
from functools import partial
from os import makedirs
from typing import Any, Callable, Literal, NamedTuple, Optional, TypeVar, Union

import jax
import numpy as np
from jax import numpy as jnp
from jax import random
from jax.tree_util import Partial, tree_map

from . import optimize
from .evi import (
    Samples, _parse_jit, draw_linear_residual, nonlinearly_update_residual
)
from .likelihood import Likelihood
from .logger import logger
from .minisanity import minisanity
from .model import LazyModel
from .tree_math import get_map, hide_strings, vdot

P = TypeVar("P")


def get_status_message(
    samples, state, residual=None, *, name="", map="lmap"
) -> str:
    energy = state.minimization_state.fun
    msg_smpl = ""
    if isinstance(state.sample_state, optimize.OptimizeResults):
        nlsi = tuple(int(el) for el in state.sample_state.nit)
        msg_smpl = f"\n{name}: #(Nonlinear sampling steps) {nlsi}"
    elif isinstance(state.sample_state, (np.ndarray, jax.Array)):
        nlsi = tuple(int(el) for el in state.sample_state)
        msg_smpl = f"\n{name}: Linear sampling status {nlsi}"
    mini_res = ""
    if residual is not None:
        _, mini_res = minisanity(samples, residual, map=map)
    _, mini_pr = minisanity(samples, map=map)
    msg = (
        f"{name}: Iteration {state.nit:04d} ⛰:{energy:+2.4e}"
        f"{msg_smpl}"
        f"\n{name}: #(KL minimization steps) {state.minimization_state.nit}"
        f"\n{name}: Likelihood residual(s):\n{mini_res}"
        f"\n{name}: Prior residual(s):\n{mini_pr}"
        f"\n"
    )
    return msg


_reduce = partial(tree_map, partial(jnp.mean, axis=0))


class _StandardHamiltonian(LazyModel):
    """Joined object storage composed of a user-defined likelihood and a
    standard normal prior.
    """
    likelihood: Likelihood = field(metadata=dict(static=False))

    def __init__(self, likelihood: Likelihood, /):
        self.likelihood = likelihood

    def __call__(self, primals, **primals_kw):
        return self.energy(primals, **primals_kw)

    def energy(self, primals, **primals_kw):
        return self.likelihood(primals, **
                               primals_kw) + 0.5 * vdot(primals, primals)

    def metric(self, primals, tangents, **primals_kw):
        return self.likelihood.metric(
            primals, tangents, **primals_kw
        ) + tangents


def _kl_vg(
    likelihood,
    primals,
    primals_samples,
    *,
    map=jax.vmap,
    reduce=_reduce,
):
    assert isinstance(primals_samples, Samples)
    map = get_map(map)
    ham = _StandardHamiltonian(likelihood)

    if len(primals_samples) == 0:
        return jax.value_and_grad(ham)(primals)
    vvg = map(jax.value_and_grad(ham))
    s = vvg(primals_samples.at(primals).samples)
    return reduce(s)


def _kl_met(
    likelihood,
    primals,
    tangents,
    primals_samples,
    *,
    map=jax.vmap,
    reduce=_reduce
):
    assert isinstance(primals_samples, Samples)
    map = get_map(map)
    ham = _StandardHamiltonian(likelihood)

    if len(primals_samples) == 0:
        return ham.metric(primals, tangents)
    vmet = map(ham.metric, in_axes=(0, None))
    s = vmet(primals_samples.at(primals).samples, tangents)
    return reduce(s)


@jax.jit
def concatenate_zip(*arrays):
    return tree_map(
        lambda *x: jnp.stack(x, axis=1).reshape((-1, ) + x[0].shape[1:]),
        *arrays
    )


SMPL_MODE_TYP = Literal[
    "linear_sample",
    "linear_resample",
    "nonlinear_sample",
    "nonlinear_resample",
    "nonlinear_update",
]
SMPL_MODE_GENERIC_TYP = Union[SMPL_MODE_TYP, Callable[[int], SMPL_MODE_TYP]]
DICT_OR_CALL4DICT_TYP = Union[Callable[[int], dict], dict]


class OptimizeVIState(NamedTuple):
    nit: int
    key: Any
    sample_state: Optional[optimize.OptimizeResults] = None
    minimization_state: Optional[optimize.OptimizeResults] = None
    config: dict[str, Union[dict, Callable[[int], Any], Any]] = {}


def _getitem_at_nit(config, key, nit):
    c = config[key]
    if callable(c) and len(inspect.getfullargspec(c).args) == 1:
        return c(nit)
    return c


class OptimizeVI:
    """State-less assembly of all methods needed for an MGVI/geoVI style VI
    approximation.

    Builds functions for a VI approximation via variants of the `Geometric
    Variational Inference` and/or `Metric Gaussian Variational Inference`
    algorithms. They produce approximate posterior samples that are used for KL
    estimation internally and the final set of samples are the approximation of
    the posterior. The samples can be linear, i.e. following a standard normal
    distribution in model space, or nonlinear, i.e. following a standard normal
    distribution in the canonical coordinate system of the Riemannian manifold
    associated with the metric of the approximate posterior distribution. The
    coordinate transformation for the nonlinear sample is approximated by an
    expansion.

    Both linear and nonlinear sample start by drawing a sample from the
    inverse metric. To do so, we draw a sample which has the metric as
    covariance structure and apply the inverse metric to it. The sample
    transformed in this way has the inverse metric as covariance. The first
    part is trivial since we can use the left square root of the metric
    :math:`L` associated with every likelihood:

    .. math::
        \\tilde{d} \\leftarrow \\mathcal{G}(0,\\mathbb{1}) \\
        t = L \\tilde{d}

    with :math:`t` now having a covariance structure of

    .. math::
        <t t^\\dagger> = L <\\tilde{d} \\tilde{d}^\\dagger> L^\\dagger = M .

    To transform the sample to an inverse sample, we apply the inverse
    metric. We can do so using the conjugate gradient algorithm (CG). The CG
    algorithm yields the solution to :math:`M s = t`, i.e. applies the
    inverse of :math:`M` to :math:`t`:

    .. math::
        M &s =  t \\\\
        &s = M^{-1} t = cg(M, t) .

    The linear sample is :math:`s`.

    The nonlinear sampling uses :math:`s` as a starting value and curves it in
    a nonlinear way as to better resemble the posterior locally. See the below
    reference literature for more details on the nonlinear sampling.

    See also
    --------
    `Geometric Variational Inference`, Philipp Frank, Reimar Leike,
    Torsten A. Enßlin, `<https://arxiv.org/abs/2105.10470>`_
    `<https://doi.org/10.3390/e23070853>`_
    `Metric Gaussian Variational Inference`, Jakob Knollmüller,
    Torsten A. Enßlin, `<https://arxiv.org/abs/1901.11033>`_
    """
    def __init__(
        self,
        likelihood: Likelihood,
        n_total_iterations: int,
        *,
        kl_jit=True,
        residual_jit=True,
        kl_map=jax.vmap,
        residual_map="lmap",
        kl_reduce=_reduce,
        mirror_samples=True,
        _kl_value_and_grad: Optional[Callable] = None,
        _kl_metric: Optional[Callable] = None,
        _draw_linear_residual: Optional[Callable] = None,
        _nonlinearly_update_residual: Optional[Callable] = None,
        _get_status_message: Optional[Callable] = None,
    ):
        """JaxOpt style minimizer for a VI approximation of a distribution with
        samples.

        Parameters
        ----------
        likelihood: :class:`~nifty8.re.likelihood.Likelihood`
            Likelihood to be used for inference.
        n_total_iterations: int
            Total number of iterations. One iteration consists of the steps
            1) - 3).
        kl_jit: bool or callable
            Whether to jit the KL minimization.
        residual_jit: bool or callable
            Whether to jit the residual sampling functions.
        kl_map: callable or str
            Map function used for the KL minimization.
        residual_map: callable or str
            Map function used for the residual sampling functions.
        kl_reduce: callable
            Reduce function used for the KL minimization.
        mirror_samples: bool
            Whether to mirror the samples or not.

        Notes
        -----
        Implements the base logic present in conditional VI approximations
        such as MGVI and geoVI. First samples are generated (and/or updated)
        and then their collective mean is optimized for using the sample
        estimated variational KL between the true distribution and the sampled
        approximation. This is split into three steps:
        1) Sample generation
        2) Sample update
        3) KL minimization.
        Step 1) and 2) may be skipped depending on the minimizers state, but
        step 3) is always performed at the end of one iteration. A full loop
        consists of repeatedly iterating over the steps 1) - 3).
        """
        kl_jit = _parse_jit(kl_jit)
        residual_jit = _parse_jit(residual_jit)
        residual_map = get_map(residual_map)

        if mirror_samples is False:
            raise NotImplementedError()

        if _kl_value_and_grad is None:
            _kl_value_and_grad = partial(
                kl_jit(_kl_vg, static_argnames=("map", "reduce")),
                likelihood,
                map=kl_map,
                reduce=kl_reduce
            )
        if _kl_metric is None:
            _kl_metric = partial(
                kl_jit(_kl_met, static_argnames=("map", "reduce")),
                likelihood,
                map=kl_map,
                reduce=kl_reduce
            )
        if _draw_linear_residual is None:
            _draw_linear_residual = partial(
                residual_jit(draw_linear_residual), likelihood
            )
        if _nonlinearly_update_residual is None:
            # TODO: Pull out `jit` from `nonlinearly_update_residual` once NCG
            # is jit-able
            from .evi import _nonlinearly_update_residual_functions

            _nonlin_funcs = _nonlinearly_update_residual_functions(
                likelihood=likelihood,
                jit=residual_jit,
            )
            _nonlinearly_update_residual = partial(
                nonlinearly_update_residual,
                None, # Explicify no likelihood dependency
                _nonlinear_update_funcs=_nonlin_funcs,
            )
        if _get_status_message is None:
            _get_status_message = partial(
                get_status_message,
                residual=likelihood.normalized_residual,
                name=self.__class__.__name__,
            )

        self.n_total_iterations = n_total_iterations
        self.kl_value_and_grad = _kl_value_and_grad
        self.kl_metric = _kl_metric
        self.draw_linear_residual = _draw_linear_residual
        self.nonlinearly_update_residual = _nonlinearly_update_residual
        self.residual_map = residual_map
        self.get_status_message = _get_status_message

    def draw_linear_samples(self, primals, keys, **kwargs):
        # NOTE, use `Partial` in favor of `partial` to allow the (potentially)
        # re-jitting `residual_map` to trace the kwargs
        kwargs = hide_strings(kwargs)
        sampler = Partial(self.draw_linear_residual, **kwargs)
        sampler = self.residual_map(sampler, in_axes=(None, 0))
        smpls, smpls_states = sampler(primals, keys)
        # zip samples such that the mirrored-counterpart always comes right
        # after the original sample
        smpls = Samples(
            pos=primals, samples=concatenate_zip(smpls, -smpls), keys=keys
        )
        return smpls, smpls_states

    def nonlinearly_update_samples(self, samples: Samples, **kwargs):
        # NOTE, use `Partial` in favor of `partial` to allow the (potentially)
        # re-jitting `residual_map` to trace the kwargs
        kwargs = hide_strings(kwargs)
        curver = Partial(self.nonlinearly_update_residual, **kwargs)
        curver = self.residual_map(curver, in_axes=(None, 0, 0, 0))
        assert len(samples.keys) == len(samples) // 2
        metric_sample_key = concatenate_zip(*((samples.keys, ) * 2))
        sgn = jnp.ones(len(samples.keys))
        sgn = concatenate_zip(sgn, -sgn)
        smpls, smpls_states = curver(
            samples.pos, samples._samples, metric_sample_key, sgn
        )
        smpls = Samples(pos=samples.pos, samples=smpls, keys=samples.keys)
        return smpls, smpls_states

    def draw_samples(
        self,
        samples: Samples,
        *,
        key,
        sample_mode: SMPL_MODE_TYP,
        n_samples: int,
        point_estimates,
        draw_linear_kwargs={},
        nonlinearly_update_kwargs={},
        **kwargs
    ):
        # Always resample if `n_samples` increased
        n_keys = 0 if samples.keys is None else len(samples.keys)
        if n_samples == 0:
            sample_mode = ""
        elif n_samples != n_keys and sample_mode.lower() == "nonlinear_update":
            sample_mode = "nonlinear_resample"
        elif n_samples != n_keys and sample_mode.lower().endswith("_sample"):
            sample_mode = sample_mode.replace("_sample", "_resample")

        if sample_mode.lower() in (
            "linear_resample", "linear_sample", "nonlinear_resample",
            "nonlinear_sample"
        ):
            k_smpls = samples.keys  # Re-use the keys if not re-sampling
            if sample_mode.lower().endswith("_resample"):
                k_smpls = random.split(key, n_samples)
            assert n_samples == len(k_smpls)
            samples, st_smpls = self.draw_linear_samples(
                samples.pos,
                k_smpls,
                point_estimates=point_estimates,
                **draw_linear_kwargs,
                **kwargs
            )
            if sample_mode.lower().startswith("nonlinear"):
                samples, st_smpls = self.nonlinearly_update_samples(
                    samples,
                    point_estimates=point_estimates,
                    **nonlinearly_update_kwargs,
                    **kwargs
                )
            elif not sample_mode.lower().startswith("linear"):
                ve = f"invalid sampling mode {sample_mode!r}"
                raise ValueError(ve)
        elif sample_mode.lower() == "nonlinear_update":
            samples, st_smpls = self.nonlinearly_update_samples(
                samples,
                point_estimates=point_estimates,
                **nonlinearly_update_kwargs,
                **kwargs
            )
        elif sample_mode == "":
            samples, st_smpls = samples, 0  # Do nothing for MAP
        else:
            ve = f"invalid sampling mode {sample_mode!r}"
            raise ValueError(ve)
        return samples, st_smpls

    def kl_minimize(
        self,
        samples: Samples,
        minimize: Callable[..., optimize.OptimizeResults] = optimize._newton_cg,
        minimize_kwargs={},
        **kwargs
    ) -> optimize.OptimizeResults:
        fun_and_grad = Partial(
            self.kl_value_and_grad, primals_samples=samples, **kwargs
        )
        hessp = Partial(self.kl_metric, primals_samples=samples, **kwargs)
        kl_opt_state = minimize(
            None,
            x0=samples.pos,
            fun_and_grad=fun_and_grad,
            hessp=hessp,
            **minimize_kwargs
        )
        return kl_opt_state

    def init_state(
        self,
        key,
        *,
        nit=0,
        n_samples: Union[int, Callable[[int], int]],
        draw_linear_kwargs: DICT_OR_CALL4DICT_TYP = dict(
            cg_name="SL", cg_kwargs=dict()
        ),
        nonlinearly_update_kwargs: DICT_OR_CALL4DICT_TYP = dict(
            minimize_kwargs=dict(name="SN", cg_kwargs=dict(name="SNCG"))
        ),
        kl_kwargs: DICT_OR_CALL4DICT_TYP = dict(
            minimize_kwargs=dict(name="M", cg_kwargs=dict(name="MCG"))
        ),
        sample_mode: SMPL_MODE_GENERIC_TYP = "nonlinear_resample",
        point_estimates=(),
        constants=(),  # TODO
    ) -> OptimizeVIState:
        """Initialize the state of the (otherwise state-less) VI approximation.

        Parameters
        ----------
        key : jax random number generataion key
        nit : int
            Current iteration number.
        n_samples : int or callable
            Number of samples to draw.
        draw_linear_kwargs : dict or callable
            Configuration for drawing linear samples, see
            :func:`draw_linear_residual`.
        nonlinearly_update_kwargs : dict or callable
            Configuration for nonlinearly updating samples, see
            :func:`nonlinearly_update_residual`.
        kl_kwargs : dict or callable
            Keyword arguments for the KL minimizer.
        sample_mode : str or callable
            One in {"linear_sample", "linear_resample", "nonlinear_sample",
            "nonlinear_resample", "nonlinear_update"}. The mode denotes the way
            samples are drawn and/or updates, "linear" draws MGVI samples,
            "nonlinear" draws MGVI samples which are then nonlinearly updated
            with geoVI, the "_sample" versus "_resample" suffix denotes whether
            the same stochasticity or new stochasticity is used for the drawing
            of the samples, and "nonlinear_update" nonlinearly updates existing
            samples using geoVI.
        point_estimates: tree-like structure or tuple of str
            Pytree of same structure as likelihood input but with boolean
            leaves indicating whether to sample the value in the input or use
            it as a point estimate. As a convenience method, for dict-like
            inputs, a tuple of strings is also valid. From these the boolean
            indicator pytree is automatically constructed.
        constants: tree-like structure or tuple of str
            Not implemented yet, sorry :( Do bug me (Gordian) at
            edh@mpa-garching.mpg.de if you wanted to run with this option.

        Most of the parameters can be callable, in which case they are called
        with the current iteration number as argument and should return the
        value to use for the current iteration.
        """
        config = dict(
            n_samples=n_samples,
            sample_mode=sample_mode,
            point_estimates=point_estimates,
            constants=constants,
            draw_linear_kwargs=draw_linear_kwargs,
            nonlinearly_update_kwargs=nonlinearly_update_kwargs,
            kl_kwargs=kl_kwargs,
        )
        return OptimizeVIState(nit, key, config=config)

    def update(
        self,
        samples: Samples,
        state: OptimizeVIState,
        /,
        **kwargs,
    ) -> tuple[Samples, OptimizeVIState]:
        """Moves the VI approximation one sample update and minimization forward.

        Parameters
        ----------
        samples : :class:`Samples`
            Current samples.
        state : :class:`OptimizeVIState`
            Current state of the VI approximation.
        kwargs : dict
            Keyword arguments passed to the residual sampling functions.
        """
        assert isinstance(samples, Samples)
        assert isinstance(state, OptimizeVIState)
        nit, key, config = state.nit, state.key, state.config

        constants = _getitem_at_nit(config, "constants", nit)
        if not (constants == () or constants is None):
            raise NotImplementedError()

        sample_mode = _getitem_at_nit(config, "sample_mode", nit)
        point_estimates = _getitem_at_nit(config, "point_estimates", nit)
        n_samples = _getitem_at_nit(config, "n_samples", nit)
        draw_linear_kwargs = _getitem_at_nit(config, "draw_linear_kwargs", nit)
        nonlinearly_update_kwargs = _getitem_at_nit(
            config, "nonlinearly_update_kwargs", nit
        )
        # Make the `key` tick independently of whether samples are drawn or not
        key, sk = random.split(key, 2)
        samples, st_smpls = self.draw_samples(
            samples,
            key=sk,
            sample_mode=sample_mode,
            point_estimates=point_estimates,
            n_samples=n_samples,
            draw_linear_kwargs=draw_linear_kwargs,
            nonlinearly_update_kwargs=nonlinearly_update_kwargs,
            **kwargs
        )

        kl_kwargs = _getitem_at_nit(config, "kl_kwargs", nit).copy()
        kl_opt_state = self.kl_minimize(samples, **kl_kwargs, **kwargs)
        samples = samples.at(kl_opt_state.x)
        # Remove unnecessary references
        kl_opt_state = kl_opt_state._replace(
            x=None, jac=None, hess=None, hess_inv=None
        )

        state = state._replace(
            nit=nit + 1,
            key=key,
            sample_state=st_smpls,
            minimization_state=kl_opt_state,
        )
        return samples, state

    def run(self, samples, *args, **kwargs) -> tuple[Samples, OptimizeVIState]:
        state = self.init_state(*args, **kwargs)
        nm = self.__class__.__name__
        for i in range(state.nit, self.n_total_iterations):
            logger.info(f"{nm}: Starting {i+1:04d}")
            samples, state = self.update(samples, state)
            msg = self.get_status_message(
                samples, state, map=self.residual_map, name=nm
            )
            logger.info(msg)
        return samples, state


def optimize_kl(
    likelihood: Likelihood,
    position_or_samples,
    *,
    key,
    n_total_iterations: int,
    n_samples,
    point_estimates=(),
    constants=(),
    kl_jit=True,
    residual_jit=True,
    kl_map=jax.vmap,
    residual_map="lmap",
    kl_reduce=_reduce,
    mirror_samples=True,
    draw_linear_kwargs=dict(cg_name="SL", cg_kwargs=dict()),
    nonlinearly_update_kwargs=dict(
        minimize_kwargs=dict(name="SN", cg_kwargs=dict(name="SNCG"))
    ),
    kl_kwargs=dict(minimize_kwargs=dict(name="M", cg_kwargs=dict(name="MCG"))),
    sample_mode: SMPL_MODE_GENERIC_TYP = "nonlinear_resample",
    resume: Union[str, bool] = False,
    callback: Optional[Callable[[Samples, OptimizeVIState], None]] = None,
    odir: Optional[str] = None,
    _optimize_vi=None,
    _optimize_vi_state=None,
) -> tuple[Samples, OptimizeVIState]:
    """One-stop-shop for MGVI/geoVI style VI approximation.

    Parameters
    ----------
    position_or_samples: Samples or tree-like
        Initial position for minimization.
    resume : str or bool
        Resume partially run optimization. If `True`, the optimization is
        resumed from the previos state in `odir` otherwise it is resumed from
        the location toward which `resume` points.
    callback : callable or None
        Function called after every global iteration taking the samples and the
        optimization state.
    odir : str or None
        Path at which all output files are saved.


    See :class:`OptimizeVI` and :func:`OptimizeVI.init_state` for the remaining
    parameters and further details on the optimization.
    """
    LAST_FILENAME = "last.pkl"
    MINISANITY_FILENAME = "minisanity.txt"

    opt_vi = _optimize_vi if _optimize_vi is not None else None
    if opt_vi is None:
        opt_vi = OptimizeVI(
            likelihood,
            n_total_iterations=n_total_iterations,
            kl_jit=kl_jit,
            residual_jit=residual_jit,
            kl_map=kl_map,
            residual_map=residual_map,
            kl_reduce=kl_reduce,
            mirror_samples=mirror_samples
        )

    last_fn = os.path.join(odir, LAST_FILENAME) if odir is not None else None
    resume_fn = resume if os.path.isfile(resume) else last_fn
    sanity_fn = os.path.join(
        odir, MINISANITY_FILENAME
    ) if odir is not None else None

    samples = None
    if isinstance(position_or_samples, Samples):
        samples = position_or_samples
    else:
        samples = Samples(pos=position_or_samples, samples=None, keys=None)
    opt_vi_st = None
    if resume:
        if not os.path.isfile(resume_fn):
            raise ValueError(f"unable to resume from {resume_fn!r}")
        if samples.pos is not None:
            logger.warning("overwriting `position_or_samples` with `resume`")
        with open(resume_fn, "rb") as f:
            samples, opt_vi_st = pickle.load(f)
    opt_vi_st_init = opt_vi.init_state(
        key,
        n_samples=n_samples,
        draw_linear_kwargs=draw_linear_kwargs,
        nonlinearly_update_kwargs=nonlinearly_update_kwargs,
        kl_kwargs=kl_kwargs,
        sample_mode=sample_mode,
        point_estimates=point_estimates,
        constants=constants,
    )
    opt_vi_st = _optimize_vi_state if _optimize_vi_state is not None else opt_vi_st
    opt_vi_st = opt_vi_st_init if opt_vi_st is None else opt_vi_st
    if len(opt_vi_st.config) == 0:  # resume or _optimize_vi_state has empty config
        opt_vi_st = opt_vi_st._replace(config=opt_vi_st_init.config)

    if odir:
        makedirs(odir, exist_ok=True)
    if not resume and sanity_fn is not None:
        with open(sanity_fn, "w"):
            pass
    if not resume and last_fn is not None:
        with open(last_fn, "wb"):
            pass

    nm = "OPTIMIZE_KL"
    for i in range(opt_vi_st.nit, opt_vi.n_total_iterations):
        logger.info(f"{nm}: Starting {i+1:04d}")
        samples, opt_vi_st = opt_vi.update(samples, opt_vi_st)
        msg = opt_vi.get_status_message(samples, opt_vi_st, name=nm)
        logger.info(msg)
        if sanity_fn is not None:
            with open(sanity_fn, "a") as f:
                f.write("\n" + msg)
        if last_fn is not None:
            with open(last_fn, "wb") as f:
                # TODO: Make all arrays numpy arrays as to not instantiate on
                # the main device when loading
                pickle.dump((samples, opt_vi_st._replace(config={})), f)
        if callback is not None:
            callback(samples, opt_vi_st)

    return samples, opt_vi_st



#!/usr/bin/env python3

import pytest

pytest.importorskip("jax")

from functools import partial, reduce

import jax
import jax.numpy as jnp
from jax import random
from jax.tree_util import tree_map
from numpy.testing import assert_allclose

import nifty8.re as jft

pmp = pytest.mark.parametrize


class _OverwriteableLikelihood(jft.Likelihood):
    def __init__(
        self,
        energy=None,
        *,
        metric=None,
        left_sqrt_metric=None,
        right_sqrt_metric=None,
        transformation=None,
        domain=None,
        lsm_tangents_shape=None,
    ):
        self._energy = energy
        self._metric = metric
        self._left_sqrt_metric = left_sqrt_metric
        self._right_sqrt_metric = right_sqrt_metric
        self._transformation = transformation
        super().__init__(domain=domain, lsm_tangents_shape=lsm_tangents_shape)

    def energy(self, *a, **k):
        if self._energy is not None:
            return self._energy(*a, **k)
        return super().energy(*a, **k)

    def metric(self, *a, **k):
        if self._metric is not None:
            return self._metric(*a, **k)
        return super().metric(*a, **k)

    def left_sqrt_metric(self, *a, **k):
        if self._left_sqrt_metric is not None:
            return self._left_sqrt_metric(*a, **k)
        return super().left_sqrt_metric(*a, **k)

    def right_sqrt_metric(self, *a, **k):
        if self._right_sqrt_metric is not None:
            return self._right_sqrt_metric(*a, **k)
        return super().right_sqrt_metric(*a, **k)

    def transformation(self, *a, **k):
        if self._transformation is not None:
            return self._transformation(*a, **k)
        return super().transformation(*a, **k)


def lst2fixt(lst):
    @pytest.fixture(params=lst)
    def fixt(request):
        return request.param

    return fixt


def random_draw(key, shape, dtype, method):
    def _isleaf(x):
        if isinstance(x, tuple):
            return bool(
                reduce(lambda a, b: a * b, (isinstance(ii, int) for ii in x))
            )
        return False

    swd = tree_map(
        lambda x: jft.ShapeWithDtype(x, dtype), shape, is_leaf=_isleaf
    )
    return jft.random_like(key, jft.Vector(swd), method)


def random_noise_std_inv(key, shape):
    diag = 1. / random_draw(key, shape, float, random.exponential)

    def noise_std_inv(tangents):
        return diag * tangents

    return noise_std_inv


seed = lst2fixt((3639, 12, 41, 42))
shape = lst2fixt(
    ((4, 2), (2, 1), (5, ), [(2, 3), (1, 2)], ((2, ), {
        'a': (3, 1)
    }))
)

lh_init_true = (
    (
        jft.Gaussian, {
            "data": partial(random_draw, dtype=float, method=random.normal),
            "noise_std_inv": random_noise_std_inv
        }, partial(random_draw, dtype=float, method=random.normal)
    ), (
        jft.StudentT, {
            "data": partial(random_draw, dtype=float, method=random.normal),
            "dof": partial(random_draw, dtype=float, method=random.exponential),
            "noise_std_inv": random_noise_std_inv
        }, partial(random_draw, dtype=float, method=random.normal)
    ), (
        jft.Poissonian, {
            "data":
                partial(
                    random_draw,
                    dtype=int,
                    method=partial(random.poisson, lam=3.14)
                ),
        }, partial(random_draw, dtype=float, method=random.exponential)
    )
)
lh_init_approx = (
    (
        jft.VariableCovarianceGaussian, {
            "data": partial(random_draw, dtype=float, method=random.normal),
        }, lambda key, shape: (
            random_draw(key, shape, float, random.normal), 3. + 1. /
            tree_map(jnp.exp, random_draw(key, shape, float, random.normal))
        )
    ), (
        jft.VariableCovarianceStudentT, {
            "data": partial(random_draw, dtype=float, method=random.normal),
            "dof": partial(random_draw, dtype=float, method=random.exponential),
        }, lambda key, shape: (
            random_draw(key, shape, float, random.normal),
            tree_map(
                jnp.exp, 3. + 1e-1 *
                random_draw(key, shape, float, random.normal)
            )
        )
    )
)


def test_gaussian_vs_vcgaussian_consistency(seed, shape):
    rtol = 10 * jnp.finfo(jnp.zeros(0).dtype).eps
    atol = 5 * jnp.finfo(jnp.zeros(0).dtype).eps

    key = random.PRNGKey(seed)
    sk = list(random.split(key, 5))
    d = random_draw(sk.pop(), shape, float, random.normal)
    m1 = random_draw(sk.pop(), shape, float, random.normal)
    m2 = random_draw(sk.pop(), shape, float, random.normal)
    t = random_draw(sk.pop(), shape, float, random.normal)
    inv_std = random_draw(sk.pop(), shape, float, random.normal)
    inv_std = 1. / tree_map(jnp.exp, 1. + inv_std)

    gauss = jft.Gaussian(d, noise_std_inv=lambda x: inv_std * x)
    vcgauss = jft.VariableCovarianceGaussian(d)

    diff_g = gauss(m2) - gauss(m1)
    diff_vcg = vcgauss((m2, inv_std)) - vcgauss((m1, inv_std))

    tree_map(partial(assert_allclose, rtol=rtol, atol=atol), diff_g, diff_vcg)

    met_g = gauss.metric(m1, t)
    met_vcg = vcgauss.metric((m1, inv_std), (t, d / 2))[0]
    tree_map(partial(assert_allclose, rtol=rtol, atol=atol), met_g, met_vcg)


def test_studt_vs_vcstudt_consistency(seed, shape):
    rtol = 10 * jnp.finfo(jnp.zeros(0).dtype).eps
    atol = 4 * jnp.finfo(jnp.zeros(0).dtype).eps

    key = random.PRNGKey(seed)
    sk = list(random.split(key, 6))
    d = random_draw(sk.pop(), shape, float, random.normal)
    dof = random_draw(sk.pop(), shape, float, random.normal)
    m1 = random_draw(sk.pop(), shape, float, random.normal)
    m2 = random_draw(sk.pop(), shape, float, random.normal)
    t = random_draw(sk.pop(), shape, float, random.normal)
    inv_std = random_draw(sk.pop(), shape, float, random.normal)
    inv_std = 1. / tree_map(jnp.exp, 1. + inv_std)

    studt = jft.StudentT(d, dof, noise_std_inv=lambda x: inv_std * x)
    vcstudt = jft.VariableCovarianceStudentT(d, dof)

    diff_t = studt(m2) - studt(m1)
    diff_vct = vcstudt((m2, 1. / inv_std)) - vcstudt((m1, 1. / inv_std))
    tree_map(partial(assert_allclose, rtol=rtol, atol=atol), diff_t, diff_vct)

    met_t = studt.metric(m1, t)
    met_vct = vcstudt.metric((m1, 1. / inv_std), (t, d / 2))[0]
    tree_map(partial(assert_allclose, rtol=rtol, atol=atol), met_t, met_vct)


@pmp("lh_init", lh_init_true + lh_init_approx)
def test_sqrt_metric_vs_metric_consistency(seed, shape, lh_init):
    rtol = 4 * jnp.finfo(jnp.zeros(0).dtype).eps
    atol = 0.
    aallclose = partial(assert_allclose, rtol=rtol, atol=atol)

    N_TRIES = 5

    lh_init_method, draw, latent_init = lh_init
    key = random.PRNGKey(seed)
    key, *subkeys = random.split(key, 1 + len(draw))
    init_kwargs = {
        k: method(key=sk, shape=shape)
        for (k, method), sk in zip(draw.items(), subkeys)
    }
    lh = lh_init_method(**init_kwargs)

    # Let NIFTy.re infer the metric from the left-square-root-metric
    lh_mini1 = _OverwriteableLikelihood(
        lh.energy,
        left_sqrt_metric=lh.left_sqrt_metric,
        lsm_tangents_shape=lh.lsm_tangents_shape,
    )

    for _ in range(N_TRIES):
        key, *sk = random.split(key, 3)
        p = latent_init(sk.pop(), shape=shape)
        t = latent_init(sk.pop(), shape=shape)
        tree_map(aallclose, lh.metric(p, t), lh_mini1.metric(p, t))

    # Let NIFTy.re infer the metric from the left-square-root-metric and
    # right-sqrt-metric
    lh_mini2 = _OverwriteableLikelihood(
        lh.energy,
        left_sqrt_metric=lh.left_sqrt_metric,
        right_sqrt_metric=lh.right_sqrt_metric,
        lsm_tangents_shape=lh.lsm_tangents_shape,
    )

    for _ in range(N_TRIES):
        key, *sk = random.split(key, 3)
        p = latent_init(sk.pop(), shape=shape)
        t = latent_init(sk.pop(), shape=shape)
        tree_map(aallclose, lh.metric(p, t), lh_mini2.metric(p, t))


@pmp("lh_init", lh_init_true)
def test_transformation_vs_sqrt_metric_consistency(seed, shape, lh_init):
    rtol = 4 * jnp.finfo(jnp.zeros(0).dtype).eps
    atol = 0.
    aallclose = partial(assert_allclose, rtol=rtol, atol=atol)

    N_TRIES = 5

    lh_init_method, draw, latent_init = lh_init
    key = random.PRNGKey(seed)
    key, *subkeys = random.split(key, 1 + len(draw))
    init_kwargs = {
        k: method(key=sk, shape=shape)
        for (k, method), sk in zip(draw.items(), subkeys)
    }
    lh = lh_init_method(**init_kwargs)
    try:
        lh.transformation(latent_init(key, shape=shape))
    except NotImplementedError:
        pytest.skip("no transformation rule implemented yet")

    energy, lsm, lsm_shp = lh.energy, lh.left_sqrt_metric, lh.lsm_tangents_shape
    # Let NIFTy.re infer the left-square-root-metric and the metric from the
    # transformation
    lh_mini = _OverwriteableLikelihood(
        energy, left_sqrt_metric=lsm, lsm_tangents_shape=lsm_shp
    )

    for _ in range(N_TRIES):
        key, *sk = random.split(key, 3)
        p = latent_init(sk.pop(), shape=shape)
        t = latent_init(sk.pop(), shape=shape)
        tree_map(
            aallclose, lh.left_sqrt_metric(p, t),
            lh_mini.left_sqrt_metric(p, t)
        )
        tree_map(
            aallclose, lh.right_sqrt_metric(p, t),
            lh_mini.right_sqrt_metric(p, t)
        )
        tree_map(aallclose, lh.metric(p, t), lh_mini.metric(p, t))


@pmp("lh_init", lh_init_true + lh_init_approx)
def test_residuals_allfinite(seed, shape, lh_init):
    allfinite = lambda x: jnp.all(jnp.isfinite(x))

    N_TRIES = 5

    lh_init_method, draw, latent_init = lh_init
    key = random.PRNGKey(seed)
    key, *subkeys = random.split(key, 1 + len(draw))
    init_kwargs = {
        k: method(key=sk, shape=shape)
        for (k, method), sk in zip(draw.items(), subkeys)
    }
    lh = lh_init_method(**init_kwargs)

    residual = lh.normalized_residual
    for _ in range(N_TRIES):
        key, sk = random.split(key, 2)
        r = residual(latent_init(sk, shape=shape))
        assert jax.tree_util.tree_reduce(
            lambda a, b: a * b, tree_map(allfinite, r)
        )


@pmp('iscomplex', [False, True])
def test_nifty_vcgaussian_vs_niftyre_vcgaussian_consistency(seed, iscomplex):
    import nifty8 as ift

    key = random.PRNGKey(seed)

    sp = ift.RGSpace(1)
    if iscomplex:
        val = 1.0 + 1.0j
        data = ift.full(sp, 0.0 + 0.0j)
    else:
        val = 1.0
        data = ift.full(sp, 0.0)
    fl = ift.full(sp, val)
    rls = ift.Realizer(sp)
    res = ift.Adder(data, neg=True) @ ift.makeOp(fl) @ rls.adjoint
    res = res.ducktape('res')
    res = res.ducktape_left('res')
    invcov = ift.exp(ift.makeOp(ift.full(sp, 1.)))
    invcov = invcov.ducktape('invcov')
    invcov = invcov.ducktape_left('invcov')
    op = res + invcov
    dt = jnp.complex128 if iscomplex else jnp.float64
    varcov_nft = ift.VariableCovarianceGaussianEnergy(
        sp, 'res', 'invcov', dt
    ) @ op

    def op_jft(x):
        return [val * x['res'], jnp.sqrt(jnp.exp(x['invcov']))]

    varcov_jft = jft.VariableCovarianceGaussian(data.val,
                                                iscomplex).amend(op_jft)

    def random_jft_ift_field(key):
        inp = jft.random_like(key, ift.nifty2jax.convert(op.domain))
        inp_nft = ift.MultiField(
            op.domain,
            tree_map(ift.Field, op.domain.values(), tuple(inp.values()))
        )
        return inp, inp_nft

    # Test value
    key, sk = random.split(key)
    inp_jft, inp_nft = random_jft_ift_field(sk)
    lh_jft = varcov_jft(inp_jft)
    lh_nft = varcov_nft(inp_nft)
    print(f"test value :: nifty: {lh_nft.val:.4f} :: jft: {lh_jft:.4f}")
    assert_allclose(lh_nft.val, lh_jft)

    # Test metric
    key, sk = random.split(key)
    inp2_jft, inp2_nft = random_jft_ift_field(sk)
    lin = varcov_nft(ift.Linearization.make_var(inp_nft, want_metric=True))
    met_nft = lin.metric
    met_res_nft = met_nft(inp2_nft)
    met_res_jft = varcov_jft.metric(inp_jft, inp2_jft)
    print(f"test metric ::\nnifty: {met_res_nft.val}\njft: {met_res_jft}\n")
    assert_allclose(met_res_nft['invcov'].val, met_res_jft['invcov'])
    assert_allclose(met_res_nft['res'].val, met_res_jft['res'])

    # Test transform
    key, sk = random.split(key)
    inp3_jft, inp3_nft = random_jft_ift_field(sk)
    trafo_nft = varcov_nft.get_transformation()[1](inp3_nft)
    trafo_jft = varcov_jft.transformation(inp3_jft)
    print(f"test transform:\nnifty: {trafo_nft.val}\njft: {trafo_jft}\n")
    assert_allclose(trafo_nft['res'].val, trafo_jft[0])
    assert_allclose(trafo_nft['invcov'].val, trafo_jft[1])


if __name__ == "__main__":
    test_gaussian_vs_vcgaussian_consistency(42, (5, ))
